#!/usr/bin/perl

#
# Simple script to control bmpx, get info, etc.  Used Kim's MP3 Script:
# 
#   http://wiki.kwinters.org/index.php/BMPx-XChat-Perl
#
# as initial framework and some other perl related help from her.
#

use strict;
use vars qw($VERSION %IRSSI);
use vars qw($bus $service $bmpx);
use vars qw($track_uri $track_info $track_title);
use Irssi;

use utf8;
use Net::DBus;
use URI::Escape;

$VERSION = '1.08';

%IRSSI = (
    authors     => 'Martin Schlemmer',
    contact     => 'azarah@nosferatu.za.org',
    name        => 'BMPx',
    description => 'Implements /bmp command (based on XChat script by Kim).',
    license     => 'GPLv2 or later',
    changed     => '11-04-05'
);

Irssi::theme_register([
    'bmp_loaded', '%R>>%n %_Scriptinfo:%_ Loaded $0 version $1',
    'box_header', '%R,--[%n$*%R]%n',
    'box_inside', '%R|%n $*',
    'box_footer', '%R`--<%n$*%R>->%n'
]);

$bus = Net::DBus->session;

sub init_dbus ($) {
    my ($channel) = @_;

    if (! defined $service) {
        eval {
            $service = $bus->get_service("org.mpris.bmpx");
        };
        if ($@) {
            error($channel, "DBus failed to initialize.  Check that BMPx is running.");
            return 0;
        }
    }
    
    if (! defined $bmpx) {
        eval {
            $bmpx = $service->get_object("/org/mpris/bmpx/SystemControl");
        };
        if ($@) {
            error($channel, "DBus failed to initialize.  Check that BMPx is running.");
            return 0;
        }
    }
    
    return 1;
}

sub try_dbus_call($;$) {
    my ($channel, $args) = @_;
    my $retval = 0;

    if (! defined $bmpx) {
        init_dbus($channel),
            or return -1;
    }

    $retval = eval("$args");
    if ($@) {
        # Try to init DBus if we failed
        init_dbus($channel),
            or return -1;

        # And then retry.
        $retval = eval("$args");
        if ($@) {
            error($channel, "DBus operation failed.  Check that BMPx is running.");
            return -1;
        }
    }
    
    return $retval;
}

sub message ($;$) {
    my ($channel, $string) = @_;

    if ($channel) {
        $channel->print("\cB$IRSSI{name}:\cB $string", MSGLEVEL_CLIENTNOTICE);
    } else {
        Irssi::print("\cB$IRSSI{name}:\cB $string", MSGLEVEL_CLIENTNOTICE);
    }
}

sub verbose ($;$) {
    my ($channel, $string) = @_;

    message($channel, $string),
        if Irssi::settings_get_bool('bmp_verbose');
}

sub error ($;$) {
    my ($channel, $string) = @_;

    if ($channel) {
        $channel->print("\cB$IRSSI{name}:\cB $string", MSGLEVEL_CLIENTERROR);
    } else {
        Irssi::print("\cB$IRSSI{name}:\cB $string", MSGLEVEL_CLIENTERROR);
    }
}

# theme_box() is from scriptassist.pl by Stefan "tommie" Tomanek
sub theme_box ($;$;$;$) {
    my ($title, $text, $footer, $colour) = @_;

    Irssi::printformat(MSGLEVEL_CLIENTCRAP, 'box_header', $title);

    foreach (split(/\n/, $text)) {
        Irssi::printformat(MSGLEVEL_CLIENTCRAP, 'box_inside', $_);
    }

    Irssi::printformat(MSGLEVEL_CLIENTCRAP, 'box_footer', $footer);
}

sub show_help {
    my $text = '';
    my $help = "$IRSSI{name} $VERSION
/bmp
    Display info of current track
    (/ME action to curent channel)
/bmp title
    Display title of current track
    (only as a notice)
/bmp uri
    Display URI of current track
    (only as a notice)
/bmp play
    Start playback
/bmp pause
    Pause/Resume playback
/bmp stop
    Stop playback
/bmp next
    Skip to next song
/bmp prev
    Skip to previous song
/bmp vol <volume>
    Set volume to the given percentage (1-100)
    No argument display current value
";

    foreach (split(/\n/, $help)) {
        $_ =~ s/^\/(.*)$/\cB\/$1/;
        $text .= $_."\n";
    }

    theme_box("$IRSSI{name}", $text, "$IRSSI{name} help", 1);
}

sub get_size ($) {
    my $filename = shift;
    my ($filesize,@sizeinfo);
    my $size;
    
    if ($filename !~ /http:\/\// && -f "$filename") {
        $filesize = `stat -c "%s" "$filename"`;
	$filesize /= (1024*1024);
	$size = sprintf("%.2fM", $filesize);
    } else {
        $size = "N/A";
    }
    
    return $size;
}

sub get_track_info($) {
    my ($channel) = @_;

    my $index;
    $index = try_dbus_call($channel, '$bmpx->GetCurrentTrack() + 1');
    if ($index eq '-1') {
        return 0;
    }
    
    my $uri;
    $uri = try_dbus_call($channel, '$bmpx->GetCurrentUri()');
    if ($uri eq '-1') {
        return 0;
    }
    $track_uri = uri_unescape($uri);
	
    my $filename = $uri;
    $filename =~ s/file:\/\///, s/[\n\r]//;
    $filename = uri_unescape($filename);

    my $lcm;
    $lcm = try_dbus_call($channel, '$bmpx->GetMetadataForUri("'."$uri".'")');
    if ($lcm eq '-1') {
        return 0;
    }
    
    my $title = ($lcm->{artist}) ? "$lcm->{artist}" : "";
    # Strip leading/trailing spaces
    $title =~ s/^\s+//g, s/\s+$//g if $title;
    # Add Artist/Title seperator
    $title = "$title -" if $title;

    if ($lcm->{title}) {
        $title = ($title) ? "$title $lcm->{title}" : "$lcm->{title}";
	
    } else {
        my $basename = $filename;
	
	$basename =~ s/.*\///;
        $title = ($title) ? "$title $basename" : "$basename";
    }
    # Strip leading/trailing spaces
    $title =~ s/^\s+//g, s/\s+$//g if $title;
    # Strip too many spaces
    $title =~ s/\s+/ /g if $title;
    
    my $sfreq = sprintf("%.1f kHz", $lcm->{samplerate}/1000);
    my $bitrate = "$lcm->{bitrate} kbps";
    my $length = "$lcm->{time}s"; # Crashes on reload: $length = gettime($lcm->{time});
    my $size = get_size($filename);

    $track_title = $title;

    $track_info = "ME \cC14is playing ::[ \cC09$index \cC15|\cC03 $title \cC15|\cC10 $bitrate \cC15|\cC10 $sfreq\cC14 ]::";
    
    $track_info = "ME is playing ::[ $index | $title | $bitrate | $sfreq ]::",
        unless Irssi::settings_get_bool('bmp_use_color');
    
    return 1;
}

sub bmp_handler ($;$;$) {
    my ($data, $server, $channel) = @_;
    my @cmds = split /\s+/, $data;
    my $cmd = @cmds[0];
    my $cmd_data = @cmds[1];
    my $retval;

    if (! defined $bmpx) {
        init_dbus($channel),
            or return;
    }

    if (!$cmd) {
	if (!$channel) {
            error($channel, "Current window not bound to a channel.");
            return;
        }
	
        get_track_info($channel),
            or return;
        $channel->command("$track_info");

    } elsif ($cmd eq "title") {
        get_track_info($channel),
            or return;
        message($channel, "Title: $track_title"),

    } elsif ($cmd eq "uri") {
        get_track_info($channel),
            or return;
        message($channel, "URI: $track_uri"),

    } elsif ($cmd eq "play") {
        $retval = try_dbus_call($channel, '$bmpx->PlayCurrent()');
        if ($retval eq '-1') {
            return;
        }
        verbose($channel, 'Playing');

    } elsif ($cmd eq "pause") {
        $retval = try_dbus_call($channel, '$bmpx->PlayPause()');
        if ($retval eq '-1') {
            return;
        }
        verbose($channel, 'Paused');

    } elsif ($cmd eq "stop") {
        $retval = try_dbus_call($channel, '$bmpx->PlayStop()');
        if ($retval eq '-1') {
            return;
        }
        verbose($channel, 'Stopped');

    } elsif ($cmd eq "next") {
        $retval = try_dbus_call($channel, '$bmpx->PlayNext()');
        if ($retval eq '-1') {
            return;
        }
        verbose($channel, 'Next Track');

    } elsif ($cmd eq "prev") {
        $retval = try_dbus_call($channel, '$bmpx->PlayPrev()');
        if ($retval eq '-1') {
            return;
        }
        verbose($channel, 'Prev Track');

    } elsif ($cmd eq "vol") {
        my $vol = $cmd_data;

        if (!$vol) {
            $vol = try_dbus_call($channel, '$bmpx->VolumeGet()');
            if ($vol eq '-1') {
                return;
            }
            message($channel, "Volume: $vol");

        } elsif ($vol !~ /^\d+$/ || $vol < 0 || $vol > 100) {
            error($channel, 'Volume must be between 0 and 100.');
            return;

        } else {
            $retval = try_dbus_call($channel, '$bmpx->VolumeSet("'."$vol".'")');
            if ($retval eq '-1') {
                return;
            }
            verbose($channel, "Volume: $vol");
        }

    } elsif ($cmd eq "help") {
        show_help;

    } else {
        error($channel, 'Unknown command for /bmp, see /help bmp for help.');
        return;
    }
}

Irssi::settings_add_bool($IRSSI{name}, 'bmp_verbose', 1);
Irssi::settings_add_bool($IRSSI{name}, 'bmp_use_color', 1);

Irssi::command_bind("bmp", "bmp_handler");
Irssi::command_bind("bmp title", "bmp_handler");
Irssi::command_bind("bmp uri", "bmp_handler");
Irssi::command_bind("bmp play", "bmp_handler");
Irssi::command_bind("bmp pause", "bmp_handler");
Irssi::command_bind("bmp stop", "bmp_handler");
Irssi::command_bind("bmp next", "bmp_handler");
Irssi::command_bind("bmp prev", "bmp_handler");
Irssi::command_bind("bmp vol", "bmp_handler");
Irssi::command_bind("bmp help", "bmp_handler");

Irssi::printformat(MSGLEVEL_CLIENTCRAP, 'bmp_loaded', $IRSSI{name}, $VERSION);

