/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#include <stdarg.h>
#include <stdio.h>

#include "types.h"
#include "relation.h"

/**
 * HAttribute:
 * @name: name
 * @type: type ID
 *
 * Holds the complete description of an attribute.
 **/

/**
 * HAttrFunc:
 * @attr: an #HAttribute
 * @user_data: user data
 *
 * #HAttribute function.
 **/

/**
 * HTupleFunc:
 * @tuple: a #HTuple
 * @user_data: user data
 *
 * #HTuple function.
 **/

/**
 * HRelation:
 *
 * Type representing a relation.
 **/
G_DEFINE_TYPE (HRelation, h_relation, G_TYPE_OBJECT)

static void h_relation_set_property (GObject      *object,
                                     guint         property,
                                     const GValue *value,
                                     GParamSpec   *spec);
static void h_relation_get_property (GObject      *object,
                                     guint         property,
                                     GValue       *value,
                                     GParamSpec   *spec);
static void h_relation_dispose      (GObject      *object);
static void h_relation_finalize     (GObject      *object);
static void h_relation_insert_tuple (HRelation    *self,
                                     HTuple       *tuple);
static void h_relation_delete_tuple (HRelation    *self,
                                     HTuple       *tuple);

enum
{
  PROP_NONE,
  PROP_NAME,
  PROP_DESCRIPTION,
  PROP_LAST
};

enum
{
  SIGNAL_INSERT_TUPLE,
  SIGNAL_DELETE_TUPLE,
  SIGNAL_LAST
};

static guint signals[SIGNAL_LAST];

static void
h_relation_class_init (HRelationClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->set_property = h_relation_set_property;
  object_class->get_property = h_relation_get_property;
  object_class->dispose = h_relation_dispose;
  object_class->finalize = h_relation_finalize;

  klass->insert_tuple = h_relation_insert_tuple;
  klass->delete_tuple = h_relation_delete_tuple;

  g_object_class_install_property (object_class, PROP_NAME,
                                   g_param_spec_string ("name",
                                                        Q_("name"),
                                                        Q_("Name of relation"),
                                                        NULL,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_object_class_install_property (object_class, PROP_DESCRIPTION,
                                   g_param_spec_string ("description",
                                                        Q_("description"),
                                                        Q_("Description of relation"),
                                                        NULL,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));

  signals[SIGNAL_INSERT_TUPLE] = g_signal_new ("insert-tuple",
                                               H_TYPE_RELATION,
                                               G_SIGNAL_RUN_FIRST,
                                               G_STRUCT_OFFSET (HRelationClass, insert_tuple),
                                               NULL,
                                               NULL,
                                               g_cclosure_marshal_VOID__BOXED,
                                               G_TYPE_NONE,
                                               1,
                                               H_TYPE_TUPLE | G_SIGNAL_TYPE_STATIC_SCOPE);

  signals[SIGNAL_DELETE_TUPLE] = g_signal_new ("delete-tuple",
                                               H_TYPE_RELATION,
                                               G_SIGNAL_RUN_FIRST,
                                               G_STRUCT_OFFSET (HRelationClass, delete_tuple),
                                               NULL,
                                               NULL,
                                               g_cclosure_marshal_VOID__BOXED,
                                               G_TYPE_NONE,
                                               1,
                                               H_TYPE_TUPLE | G_SIGNAL_TYPE_STATIC_SCOPE);
}

static void
h_relation_init (HRelation *self)
{}

static void
h_relation_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_parent_class)->dispose (object);
}

static void
h_relation_finalize (GObject *object)
{
  HRelation *self = H_RELATION (object);

  g_free (self->description);
  g_free (self->name);

  G_OBJECT_CLASS (h_relation_parent_class)->finalize (object);
}

static void
h_relation_set_property (GObject      *object,
                         guint         property,
                         const GValue *value,
                         GParamSpec   *spec)
{
  HRelation *self = H_RELATION (object);

  switch (property)
    {
    case PROP_NAME:
      self->name = g_strdup (g_value_get_string (value));
      break;
    case PROP_DESCRIPTION:
      self->description = g_strdup (g_value_get_string (value));
      break;
    }
}

static void
h_relation_get_property (GObject    *object,
                         guint       property,
                         GValue     *value,
                         GParamSpec *spec)
{
  HRelation *self = H_RELATION (object);

  switch (property)
    {
    case PROP_NAME:
      g_value_set_string (value, self->name);
      break;
    case PROP_DESCRIPTION:
      g_value_set_string (value, self->description);
      break;
    }
}

static void
print_attr (HAttribute *attr)
{
  printf ("%s (%s)\n", attr->name, g_type_name (attr->type));
}

static void
print_tuple (HTuple *tuple)
{
  h_tuple_print (tuple);
  printf ("\n");
}

/**
 * h_relation_set_name
 * @relation: a #HRelation
 * @name: name
 *
 * Set the name of @relation.
 **/
void
h_relation_set_name (HRelation   *self,
                     const gchar *name)
{
  g_return_if_fail (H_IS_RELATION (self));

  g_object_set (G_OBJECT (self), "name", name, NULL);
}

/**
 * h_relation_set_desc:
 * @relation: a #HRelation
 * @desc: description
 *
 * Set the description of @relation.
 **/
void
h_relation_set_desc (HRelation    *self,
                     const gchar *desc)
{
  g_return_if_fail (H_IS_RELATION (self));

  g_object_set (G_OBJECT (self), "description", desc, NULL);
}

/**
 * h_relation_get_name:
 * @relation: a #HRelation
 *
 * Returns the name of @relation.
 *
 * Return value: a copy of the name string, or %NULL if no name is
 * set.
 **/
gchar *
h_relation_get_name (HRelation *self)
{
  gchar *name;

  g_return_val_if_fail (H_IS_RELATION (self), NULL);

  g_object_get (G_OBJECT (self), "name", &name, NULL);
  return name;
}

/**
 * h_relation_get_desc:
 * @relation: a #HRelation
 *
 * Returns the description of @relation.
 *
 * Return value: a copy of the description string, or %NULL if no
 * description exists
 **/
gchar *
h_relation_get_desc (HRelation *self)
{
  gchar *desc;

  g_return_val_if_fail (H_IS_RELATION (self), NULL);

  g_object_get (G_OBJECT (self), "description", &desc, NULL);
  return desc;
}

/**
 * h_relation_get_attr:
 * @relation: a #HRelation
 * @name: name of attribute
 *
 * Returns an attribute of @relation.
 *
 * Return value: attribute
 **/
HAttribute *
h_relation_get_attr (HRelation   *self,
                     const gchar *name)
{
  g_return_val_if_fail (H_IS_RELATION (self), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return H_RELATION_GET_CLASS (self)->get_attr (self, name);
}

/**
 * h_relation_print:
 * @relation: a #HRelation
 *
 * Prints @relation to stdout.
 **/
void
h_relation_print (HRelation *self)
{
  g_return_if_fail (H_IS_RELATION (self));

  printf ("Name: %s\n", self->name);

  if (self->description)
    printf ("Description: %s\n", self->description);

  H_RELATION_GET_CLASS (self)->print (self);

  printf ("Attributes: \n");
  h_relation_for_each_attr (self, (HAttrFunc) print_attr, NULL);

  /* print tuples */
  printf ("Value: \n");
  h_relation_for_each_tuple (self, (HTupleFunc) print_tuple, NULL);
}

/**
 * h_relation_for_each_attr:
 * @relation: a #HRelation
 * @func: function called for each attribute in @relation
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each attribute in @relation.
 **/
void
h_relation_for_each_attr (HRelation *self,
                          HAttrFunc  func,
                          gpointer   user_data)
{
  g_return_if_fail (H_IS_RELATION (self));

  H_RELATION_GET_CLASS (self)->for_each_attr (self, func, user_data);
}

/**
 * h_relation_for_each_tuple:
 * @relation: a #HRelation
 * @func: function called for each tuple in @relation
 * @user_data: custom data to pass to @func
 *
 * Performs a function over each tuple in @relation.
 **/
void
h_relation_for_each_tuple (HRelation  *self,
                           HTupleFunc  func,
                           gpointer    user_data)
{
  g_return_if_fail (H_IS_RELATION (self));

  H_RELATION_GET_CLASS (self)->for_each_tuple (self, func, user_data);
}

/**
 * h_relation_get_size:
 * @relation: a #HRelation
 *
 * Returns the number of tuples in @relation.
 *
 * Return value: number of tuples
 **/
guint
h_relation_get_size (HRelation *self)
{
  g_return_val_if_fail (H_IS_RELATION (self), 0);

  return H_RELATION_GET_CLASS (self)->get_size (self);
}

/**
 * h_relation_get_n_attrs:
 * @relation: a #HRelation
 *
 * Returns the number of attributes in @relation. This is an
 * alternative name for h_relation_get_degree().
 *
 * Return value: number of attributes
 **/

/**
 * h_relation_get_degree:
 * @relation: a #HRelation
 *
 * Returns the degree of @relation.
 *
 * Return value: degree
 **/
guint
h_relation_get_degree (HRelation *self)
{
  g_return_val_if_fail (H_IS_RELATION (self), 0);

  return H_RELATION_GET_CLASS (self)->get_degree (self);
}


/**
 * h_relation_ref:
 * @relation: a #HRelation
 *
 * Increases the reference count of @relation.
 **/

/**
 * h_relation_unref:
 * @relation: a #HRelation
 *
 * Decreases the reference count of @relation. @relation will be
 * destroyed when its reference count goes to 0.
 **/

static void
h_relation_insert_tuple (HRelation *self,
                         HTuple    *tuple)
{
  /* do nothing */
}

static void
h_relation_delete_tuple (HRelation *self,
                         HTuple    *tuple)
{
  /* do nothing */
}
