/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_base.h"
#include "types.h"
#include "misc.h"
#include <stdio.h>
#include <string.h>


/**
 * HRelationBase:
 *
 * A #HRelation representing a base relation.
 **/
G_DEFINE_TYPE (HRelationBase, h_relation_base, H_TYPE_RELATION)

static void  h_relation_base_dispose        (GObject      *object);
static void  h_relation_base_finalize       (GObject      *object);
static void  h_relation_base_print          (HRelation    *relation);
static HAttribute *h_relation_base_get_attr (HRelation    *relation,
                                             const gchar  *name);
static guint h_relation_base_get_degree     (HRelation    *relation);
static guint h_relation_base_get_size       (HRelation    *relation);
static void  h_relation_base_for_each_attr  (HRelation    *relation,
                                             HAttrFunc     func,
                                             gpointer      user_data);
static void  h_relation_base_for_each_tuple (HRelation    *relation,
                                             HTupleFunc    func,
                                             gpointer      user_data);

static void
h_relation_base_class_init (HRelationBaseClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  object_class->dispose = h_relation_base_dispose;
  object_class->finalize = h_relation_base_finalize;

  relation_class->print = h_relation_base_print;
  relation_class->get_attr = h_relation_base_get_attr;
  relation_class->get_degree = h_relation_base_get_degree;
  relation_class->get_size = h_relation_base_get_size;
  relation_class->for_each_attr = h_relation_base_for_each_attr;
  relation_class->for_each_tuple = h_relation_base_for_each_tuple;
}

static void
h_relation_base_init (HRelationBase *self)
{
  /* nothing to do */
}

static void
h_relation_base_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_base_parent_class)->dispose (object);
}

static void
h_relation_base_finalize (GObject *object)
{
  HRelationBase *self = H_RELATION_BASE (object);

  g_slist_foreach (self->primary_key, (GFunc) g_free, NULL);
  g_slist_free (self->primary_key);

  h_attr_map_free (self->attrs);
  h_hash_index_free (self->tuples);

  G_OBJECT_CLASS (h_relation_base_parent_class)->finalize (object);
}

static HAttribute *
h_relation_base_get_attr (HRelation   *relation,
                          const gchar *name)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_val_if_fail (self != NULL, NULL);

  return h_attr_map_get (self->attrs, name);
}

static HAttrMap *
build_attr_map (va_list *args)
{
  HAttrMap    *map;
  const gchar *name;
  GType        type;

  /* Collect attribute list into a name-attribute associative map */

  map = h_attr_map_new ();

  while (1)
    {
      name = va_arg (*args, const gchar *);
      if (!name)
        break;

      type = va_arg (*args, GType);

#ifdef ENABLE_TYPE_CHECK
      if (h_attr_map_get (map, name) || !h_type_known (type))
        {
          h_attr_map_free (map);
          return NULL;
        }
#endif

      h_attr_map_insert (map, h_attr_new (name, type));
    }

  return map;
}

#ifdef ENABLE_TYPE_CHECK
static gboolean
check_key (GHashTable *map,
           GSList     *key)
{
  if (!key)
    return FALSE;

  while (key)
    {
      if (!h_attr_map_get (map, (const gchar *) key->data))
        return FALSE;

      key = key->next;
    }

  return TRUE;
}
#endif

static GSList *
duplicate_key_in_place (GSList *key)
{
  GSList *iter;

  for (iter = key; iter; iter = iter->next)
    iter->data = g_strdup (iter->data);

  return key;
}

/**
 * h_relation_base_new:
 * @name: name
 * @description: description
 * @...: name and type of an attribute, followed by another and so on
 *   then %NULL, followed by the name of a primary key attribute,
 *   followed by another and so on, then %NULL.
 *
 * Returns a new base relation with name @name, description
 * @description, specified attributes and primary key attributes.
 *
 * Return value: a base relation or %NULL
 **/
HRelation *
h_relation_base_new (const gchar *name,
                     const gchar *description,
                     ...)
{
  HRelationBase *relation;
  GHashTable    *attrs;
  GSList        *primary_key;
  va_list        args;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (description != NULL, NULL);

  va_start (args, description);

  attrs = build_attr_map (&args);
  primary_key = h_va_list_to_slist (args);

  va_end (args);

#ifdef ENABLE_TYPE_CHECK
  if (h_attr_map_get_size (attrs) == 0 || !check_key (attrs, primary_key))
    {
      h_attr_map_free (attrs);
      g_slist_free (primary_key);
      return NULL;
    }
#endif

  relation = g_object_new (H_TYPE_RELATION_BASE,
                           "name", name,
                           "description", description,
                           NULL);

  relation->primary_key = duplicate_key_in_place (primary_key);
  relation->attrs = attrs;

  relation->tuples = h_hash_index_new_l (NULL,
                                         (GDestroyNotify) h_tuple_free,
                                         primary_key);

  return H_RELATION (relation);
}

#ifdef ENABLE_TYPE_CHECK

typedef struct
{
  HTuple   *tuple;
  gboolean  match;
}
CheckAttrParams;

static void
check_attr (const HAttribute *attr,
            CheckAttrParams  *params)
{
  HTupleElem *elem;

  if (!params->match)
    return;

  elem = h_tuple_get (params->tuple, attr->name);

  if (!elem)
    {
      params->match = FALSE;
      return;
    }

  if (elem->type != G_TYPE_NONE &&
      elem->type != attr->type)
    params->match = FALSE;
}

static gboolean
h_relation_base_check_tuple (HRelationBase *self,
                             HTuple        *tuple)
{
  CheckAttrParams params;

  if (h_tuple_get_size (tuple) !=
      h_relation_base_get_degree ((HRelation *) self))
    return FALSE;

  params.tuple = tuple;
  params.match = TRUE;

  h_relation_base_for_each_attr ((HRelation *) self,
                                 (HAttrFunc) check_attr,
                                 &params);

  return params.match;
}

#endif /* #ifdef ENABLE_TYPE_CHECK */

typedef struct
{
  HTuple *tuple;
  HTuple *copy;
}
CopyAttrParams;

static void
copy_attr (const HAttribute *attr,
           CopyAttrParams   *params)
{
  h_tuple_insert (params->copy,
                  attr->name,
                  attr->type,
                  h_tuple_get (params->tuple, attr->name)->value);
}

static void
copy_insert_tuple (HRelationBase *self,
                   HTuple        *tuple)
{
  CopyAttrParams params;

  params.tuple = tuple;
  params.copy = h_tuple_new ();

  h_relation_base_for_each_attr ((HRelation *) self,
                                 (HAttrFunc) copy_attr,
                                 &params);

  h_hash_index_insert (self->tuples, params.copy, params.copy);

  g_signal_emit_by_name (self, "insert-tuple", params.copy);
}

static HTuple *
key_new_v (HRelationBase *relation,
           va_list        args)
{
  HTuple      *key;
  gint         i, key_n_attrs;
  const gchar *name;
  gpointer     value;

#ifdef ENABLE_TYPE_CHECK
  GSList      *primary_key;
  GSList      *node;

  primary_key = h_relation_base_get_primary_key (relation);
#endif

  key_n_attrs = g_slist_length (relation->primary_key);

  key = h_tuple_new ();

  for (i = 0; i < key_n_attrs; i++)
    {
      name  = va_arg (args, const gchar *);
      value = va_arg (args, gpointer);

#ifdef ENABLE_TYPE_CHECK
      node = g_slist_find_custom (primary_key, name, (GCompareFunc) strcmp);
      if (!node)
        {
          g_message ("Key does not have attribute: %s", name);
          h_tuple_free (key);
          return NULL;
        }
#endif

      h_tuple_insert (key, name,
                      h_relation_base_get_attr ((HRelation *) relation, name)->type,
                      value);
    }

  return key;
}

/**
 * h_relation_base_insert_tuple:
 * @relation: a #HRelationBase
 * @tuple: a #HTuple to insert
 *
 * Inserts a tuple into @relation.
 *
 * Return value: TRUE if @tuple was successfully inserted, FALSE
 * otherwise
 **/
gboolean
h_relation_base_insert_tuple (HRelationBase *self,
                              HTuple        *tuple)
{
  g_return_val_if_fail (self != NULL, FALSE);

#ifdef ENABLE_TYPE_CHECK
  if (!h_relation_base_check_tuple (self, tuple))
    {
      g_message ("Invalid tuple type");
      return FALSE;
    }
#endif

  copy_insert_tuple (self, tuple);

  return TRUE;
}

/**
 * h_relation_base_insert_tuplev:
 * @relation: a #HRelationBase
 * @...: An attribute name, followed by its corresponding value,
 *   followed by another attribute name, and so on.
 *
 * Inserts a tuple into @relation.
 *
 * Return value: TRUE if tuple was successfully inserted, FALSE
 * otherwise
 **/
gboolean
h_relation_base_insert_tuplev (HRelationBase *self,
                               ...)
{
  va_list  args;
  HTuple  *tuple;

  g_return_val_if_fail (self != NULL, FALSE);

  va_start (args, self);
  tuple = h_tuple_n_new_v (FALSE,
                           h_relation_base_get_degree ((HRelation *) self),
                           args);
  va_end (args);

  copy_insert_tuple (self, tuple);
  h_tuple_free (tuple);

  return TRUE;
}

/**
 * h_relation_base_get_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Returns a tuple in @relation that has the primary key @key.
 *
 * Return value: a #HTuple storing the tuple or %NULL if no tuples have
 * the specified key.
 **/
HTuple *
h_relation_base_get_tuple (HRelationBase *self,
                           HTuple        *key)
{
  HTuple *tuple;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), FALSE);

  tuple = (HTuple *) h_hash_index_lookup (self->tuples, key);

  return tuple ? h_tuple_clone (tuple) : NULL;
}

/**
 * h_relation_base_get_tuplev:
 * @relation: a #HRelationBase
 * @...: name and value of first key attribute, followed by the second
 *   and so on, then %NULL
 *
 * Returns a tuple in @relation that has the primary key @key.
 *
 * Return value: a #HTuple storing the tuple or %NULL if no tuples have
 * the specified key.
 **/
HTuple *
h_relation_base_get_tuplev (HRelationBase *self,
                            ...)
{
  va_list  args;
  HTuple  *key;
  HTuple  *tuple;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), NULL);

  va_start (args, self);
  key = key_new_v (self, args);
  va_end (args);

  g_return_val_if_fail (key != NULL, NULL);

  tuple = h_relation_base_get_tuple (self, key);
  h_tuple_free (key);

  return tuple;
}

/**
 * h_relation_base_delete_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Deletes a tuple in @relation that has the primary key @key.
 *
 * Return value: TRUE if tuple was deleted, FALSE if no such tuple
 * exists.
 **/
gboolean
h_relation_base_delete_tuple (HRelationBase *self,
                              HTuple        *key)
{
  HTuple *tuple;

  g_return_val_if_fail (self != NULL, FALSE);

  tuple = (HTuple *) h_hash_index_lookup (self->tuples, key);
  g_return_val_if_fail (tuple != NULL, FALSE);

  g_signal_emit_by_name (self, "delete-tuple", tuple);

  return h_hash_index_remove (self->tuples, key);
}

/**
 * h_relation_base_delete_tuplev:
 * @relation: a #HRelationBase
 * @...: name and value of first key attribute, followed by the second
 *   and so on, then %NULL.
 *
 * Deletes the tuple in @relation that has the specified primary key.
 *
 * Return value: TRUE if tuple was deleted, FALSE if no such tuple
 * exists.
 **/
gboolean
h_relation_base_delete_tuplev (HRelationBase *self,
                               ...)
{
  va_list   args;
  HTuple   *key;
  gboolean  result;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), FALSE);

  va_start (args, self);
  key = key_new_v (self, args);
  va_end (args);

  g_return_val_if_fail (key != NULL, FALSE);

  result = h_relation_base_delete_tuple (self, key);
  h_tuple_free (key);

  return result;
}

/**
 * h_relation_base_get_primary_key:
 * @relation: a #HRelationBase
 *
 * Returns the primary key of @relation.
 *
 * Return value: primary key (must not be freed)
 **/
GSList *
h_relation_base_get_primary_key (HRelationBase *self)
{
  /* Should we clone the key? */

  return self->primary_key;
}

static void
h_relation_base_print (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);
  GSList *iter;

  g_return_if_fail (self != NULL);

  printf ("Primary key: ");

  for (iter = self->primary_key; iter; iter = iter->next)
    printf ("%s ", (const gchar *) iter->data);

  printf ("\n");
}

static void
h_relation_base_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_if_fail (self != NULL);

  h_attr_map_foreach (self->attrs, func, user_data);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
IterTupleParams;

static void
iter_tuple (HTuple          *key,
            HTuple          *tuple,
            IterTupleParams *params)
{
  (*params->func) (tuple, params->data);
}

static void
h_relation_base_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationBase *self = H_RELATION_BASE (relation);
  IterTupleParams params;

  g_return_if_fail (self != NULL);

  params.func = func;
  params.data = user_data;

  h_hash_index_foreach (self->tuples, (HIndexFunc) iter_tuple, &params);
}

static guint
h_relation_base_get_size (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_hash_index_get_size (self->tuples);
}

static guint
h_relation_base_get_degree (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_attr_map_get_size (self->attrs);
}
