/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include <glib-object.h>
#include "relation_project.h"
#include "misc.h"
#include <stdio.h>

/**
 * HRelationProject:
 *
 * A #HRelation representing a projected relation.
 **/
G_DEFINE_TYPE (HRelationProject, h_relation_project, H_TYPE_RELATION)

static void  h_relation_project_dispose        (GObject          *object);
static void  h_relation_project_finalize       (GObject          *object);
static void  h_relation_project_print          (HRelation        *relation);
static HAttribute *h_relation_project_get_attr (HRelation        *relation,
                                                const gchar      *name);
static guint h_relation_project_get_degree     (HRelation        *relation);
static guint h_relation_project_get_size       (HRelation        *relation);
static void  h_relation_project_for_each_attr  (HRelation        *relation,
                                                HAttrFunc         func,
                                                gpointer          user_data);
static void  h_relation_project_for_each_tuple (HRelation        *relation,
                                                HTupleFunc        func,
                                                gpointer          user_data);
static void  h_relation_project_insert_tuple   (HRelationProject *self,
                                                HTuple           *tuple);
static void  h_relation_project_delete_tuple   (HRelationProject *self,
                                                HTuple           *tuple);

static void
h_relation_project_class_init (HRelationProjectClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  object_class->dispose = h_relation_project_dispose;
  object_class->finalize = h_relation_project_finalize;

  relation_class->print = h_relation_project_print;
  relation_class->get_attr = h_relation_project_get_attr;
  relation_class->get_degree = h_relation_project_get_degree;
  relation_class->get_size = h_relation_project_get_size;
  relation_class->for_each_attr = h_relation_project_for_each_attr;
  relation_class->for_each_tuple = h_relation_project_for_each_tuple;
}

static void
h_relation_project_init (HRelationProject *self)
{
  self->to_project = NULL;
  self->attrs = NULL;
  self->index = NULL;
}

static void
h_relation_project_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_project_parent_class)->dispose (object);
}

static void
h_relation_project_finalize (GObject *object)
{
  HRelationProject *self = H_RELATION_PROJECT (object);

  g_signal_handlers_disconnect_by_func (self->to_project,
                                        (gpointer) h_relation_project_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (self->to_project,
                                        (gpointer) h_relation_project_delete_tuple,
                                        self);

  h_attr_map_free (self->attrs);
  h_hash_index_free (self->index);

  g_object_unref (self->to_project);


  G_OBJECT_CLASS (h_relation_project_parent_class)->finalize (object);
}

static void
h_relation_project_print (HRelation *relation)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);
  gchar *str;

  g_return_if_fail (self != NULL);

  printf ("Relation to project:\n");

  str = h_relation_get_name (self->to_project);
  printf ("  Name: %s\n", str);
  g_free (str);

  str = h_relation_get_desc (self->to_project);
  printf ("  Description: %s\n", str);
  g_free (str);
}

static HAttribute *
h_relation_project_get_attr (HRelation   *relation,
                             const gchar *name)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);

  g_return_val_if_fail (self != NULL, NULL);

  return h_attr_map_get (self->attrs, name);
}

static guint
h_relation_project_get_degree (HRelation *relation)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_attr_map_get_size (self->attrs);
}

static guint
h_relation_project_get_size (HRelation *relation)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_hash_index_get_size (self->index);
}

static void
h_relation_project_for_each_attr (HRelation *relation,
                                  HAttrFunc  func,
                                  gpointer   user_data)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);

  g_return_if_fail (self != NULL);

  h_attr_map_foreach (self->attrs, func, user_data);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
IterTupleParams;

static void
iter_tuple (HTuple          *key,
            HTuple          *tuple,
            IterTupleParams *params)
{
  (*params->func) (tuple, params->data);
}

static void
h_relation_project_for_each_tuple (HRelation  *relation,
                                   HTupleFunc  func,
                                   gpointer    user_data)
{
  HRelationProject *self = H_RELATION_PROJECT (relation);
  IterTupleParams   params;

  g_return_if_fail (self != NULL);

  params.func = func;
  params.data = user_data;

  h_hash_index_foreach (self->index, (HIndexFunc) iter_tuple, &params);
}

static HTuple *
insert_index (HTuple           *tuple,
              HRelationProject *self)
{
  HTuple *copy;

  copy = h_tuple_clone_elems_l (tuple,
                                h_hash_index_get_attrs (self->index));

  h_hash_index_insert (self->index, copy, copy);

  return copy;
}

static void
h_relation_project_insert_tuple (HRelationProject *self,
                                 HTuple           *tuple)
{
  if (!h_hash_index_lookup (self->index, tuple))
    {
      HTuple *new_tuple;

      new_tuple = insert_index (tuple, self);

      g_signal_emit_by_name (G_OBJECT (self), "insert-tuple", new_tuple);
    }
}

static void
h_relation_project_delete_tuple (HRelationProject *self,
                                 HTuple           *tuple)
{
  if (h_hash_index_remove (self->index, tuple))
    {
      HTuple *new_tuple;

      new_tuple = h_tuple_clone_elems_l (tuple,
                                         h_hash_index_get_attrs (self->index));

      g_signal_emit_by_name (G_OBJECT (self), "delete-tuple", new_tuple);

      h_tuple_free (new_tuple);
    }
}

static void
build_index (HRelationProject *self,
             GSList           *list)
{
  self->index = h_hash_index_new_l (NULL,
                                    (GDestroyNotify) h_tuple_free,
                                    list);

  h_relation_for_each_tuple (self->to_project,
                             (HTupleFunc) insert_index,
                             self);
}

static void
attr_map_copy_insert_attr (HAttribute *attr,
                           HAttrMap   *map)
{
  h_attr_map_insert (map, h_attr_copy (attr));
}

static HAttrMap *
attr_map_from_name_list (HRelation *relation,
                         GSList    *list)
{
  HAttrMap    *map;
  const gchar *name;

#ifdef ENABLE_TYPE_CHECK
  GSList      *iter;

  for (iter = list; iter; iter = iter->next)
    {
      name = (const gchar *) list->data;

      if (!relation_get_attr (relation, name))
        {
          g_message ("Underlying relation has no attribute '%s'", name);
          return NULL;
        }
    }
#endif

  map = h_attr_map_new ();

  if (list)
    {
      while (list)
        {
          name = (const gchar *) list->data;

          h_attr_map_insert (map,
                             h_attr_copy (h_relation_get_attr (relation, name)));

          list = list->next;
        }
    }
  else
    {
      h_relation_for_each_attr (relation,
                                (HAttrFunc) attr_map_copy_insert_attr,
                                map);
    }

  return map;
}

/**
 * h_relation_project:
 * @relation: relation to project
 * @...: name of an attribute, followed by another and so on, then a
 *   %NULL
 *
 * Creates a relation on @relation projected over a set of
 * attributes.
 *
 * Return value: projected relation
 **/
HRelation *
h_relation_project (HRelation *to_project,
                    ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  va_start (args, to_project);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_project_l (to_project, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_project_l:
 * @relation: relation to project
 * @list: list of attribute names
 *
 * A variation of h_relation_project() taking a #GSList of attribute
 * names.
 *
 * Return value: projected relation
 **/
HRelation *
h_relation_project_l (HRelation *to_project,
                      GSList    *list)
{
  HRelationProject *relation;

  g_return_val_if_fail (to_project != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_PROJECT, NULL);

  relation->to_project = to_project;
  g_object_ref (to_project);

  relation->attrs = attr_map_from_name_list (to_project, list);

  build_index (relation, list);

  g_signal_connect_swapped (G_OBJECT (to_project), "insert-tuple",
                            G_CALLBACK (h_relation_project_insert_tuple),
                            relation);

  g_signal_connect_swapped (G_OBJECT (to_project), "delete-tuple",
                            G_CALLBACK (h_relation_project_delete_tuple),
                            relation);

  return (HRelation *) relation;
}
