/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_restrict.h"
#include "misc.h"
#include <stdio.h>

/**
 * HTuplePredFunc:
 * @tuple: a #HTuple
 * @user_data: user data
 *
 * A #HTuple predicate function.
 *
 * Return value: either TRUE or FALSE
 **/

/**
 * HRelationRestrict:
 *
 * A #HRelation representing a restricted relation.
 **/
G_DEFINE_TYPE (HRelationRestrict, h_relation_restrict, H_TYPE_RELATION)

static void  h_relation_restrict_dispose        (GObject           *object);
static void  h_relation_restrict_finalize       (GObject           *object);
static void  h_relation_restrict_print          (HRelation         *relation);
static HAttribute *h_relation_restrict_get_attr (HRelation         *relation,
                                                 const gchar       *name);
static guint h_relation_restrict_get_degree     (HRelation         *relation);
static guint h_relation_restrict_get_size       (HRelation         *relation);
static void  h_relation_restrict_for_each_attr  (HRelation         *relation,
                                                 HAttrFunc          func,
                                                 gpointer           user_data);
static void  h_relation_restrict_for_each_tuple (HRelation         *relation,
                                                 HTupleFunc         func,
                                                 gpointer           user_data);
static void  h_relation_restrict_insert_tuple   (HRelationRestrict *self,
                                                 HTuple            *tuple);
static void  h_relation_restrict_delete_tuple   (HRelationRestrict *self,
                                                 HTuple            *tuple);

static void
h_relation_restrict_class_init (HRelationRestrictClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *h_relation_class = H_RELATION_CLASS (klass);

  object_class->dispose  = h_relation_restrict_dispose;
  object_class->finalize = h_relation_restrict_finalize;

  h_relation_class->print = h_relation_restrict_print;
  h_relation_class->get_attr = h_relation_restrict_get_attr;
  h_relation_class->get_degree = h_relation_restrict_get_degree;
  h_relation_class->get_size = h_relation_restrict_get_size;
  h_relation_class->for_each_attr = h_relation_restrict_for_each_attr;
  h_relation_class->for_each_tuple = h_relation_restrict_for_each_tuple;
}

static void
h_relation_restrict_init (HRelationRestrict *self)
{
  self->to_restrict = NULL;
  self->func = NULL;
  self->func_data = NULL;
}

static void
h_relation_restrict_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_restrict_parent_class)->dispose (object);
}

static void
h_relation_restrict_finalize (GObject *object)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (object);

  g_signal_handlers_disconnect_by_func (self->to_restrict,
                                        (gpointer) h_relation_restrict_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (self->to_restrict,
                                        (gpointer) h_relation_restrict_delete_tuple,
                                        self);

  h_hash_index_free (self->index);

  g_object_unref (self->to_restrict);

  G_OBJECT_CLASS (h_relation_restrict_parent_class)->finalize (object);
}

static void
h_relation_restrict_print (HRelation *relation)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);
  gchar *str;

  g_return_if_fail (self != NULL);

  printf ("Relation to restrict:\n");

  str = h_relation_get_name (self->to_restrict);
  printf ("  Name: %s\n", str ? str : "(none)");
  g_free (str);

  str = h_relation_get_desc (self->to_restrict);
  printf ("  Description: %s\n", str ? str : "(none)");
  g_free (str);
}

static HAttribute *
h_relation_restrict_get_attr (HRelation   *relation,
                              const gchar *name)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);

  g_return_val_if_fail (self != NULL, NULL);

  return h_relation_get_attr (self->to_restrict, name);
}

static guint
h_relation_restrict_get_degree (HRelation *relation)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_degree (self->to_restrict);
}

static guint
h_relation_restrict_get_size (HRelation *relation)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_hash_index_get_size (self->index);
}

static void
h_relation_restrict_for_each_attr (HRelation *relation,
                                   HAttrFunc  func,
                                   gpointer   user_data)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);

  g_return_if_fail (self != NULL);

  h_relation_for_each_attr (self->to_restrict, func, user_data);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
CallTupleFuncParams;

static void
call_tuple_func (HTuple              *tuple,
                 gpointer            *value,
                 CallTupleFuncParams *params)
{
  (*params->func) (tuple, params->data);
}

static void
h_relation_restrict_for_each_tuple (HRelation  *relation,
                                    HTupleFunc  func,
                                    gpointer    user_data)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);
  CallTupleFuncParams params;

  g_return_if_fail (self != NULL);

  params.func = func;
  params.data = user_data;

  h_hash_index_foreach (self->index, (HIndexFunc) call_tuple_func,
                        &params);
}

static void
h_relation_restrict_insert_tuple (HRelationRestrict *self,
                                  HTuple            *tuple)
{
  if ((*self->func) (tuple, self->func_data))
    {
      h_hash_index_insert (self->index, tuple, tuple);
      g_signal_emit_by_name (self, "insert-tuple", tuple);
    }
}

static void
h_relation_restrict_delete_tuple (HRelationRestrict *self,
                                  HTuple            *tuple)
{
  if (h_hash_index_remove (self->index, tuple))
    g_signal_emit_by_name (self, "delete-tuple", tuple);
}

static void
insert_tuple (HTuple            *tuple,
              HRelationRestrict *self)
{
  if ((*self->func) (tuple, self->func_data))
    h_hash_index_insert (self->index, tuple, tuple);
}

static void
init_index (HRelationRestrict *self)
{
  self->index = h_hash_index_new (NULL, NULL, NULL);

  h_relation_for_each_tuple (self->to_restrict,
                             (HTupleFunc) insert_tuple,
                             self);
}

/**
 * h_relation_restrict:
 * @relation: a #HRelation to perform restrict on
 * @predicate_func: tuple predicate function
 * @user_data: Custom data to pass to @predicate_func
 *
 * Restricts @relation to the set of tuples evaluating to TRUE with
 * @predicate_func.
 *
 * Return value: restricted #HRelation
 **/
HRelation *
h_relation_restrict (HRelation     *to_restrict,
                     HTuplePredFunc predicate_func,
                     gpointer       user_data)
{
  HRelationRestrict *relation;

  g_return_val_if_fail (to_restrict != NULL, NULL);
  g_return_val_if_fail (predicate_func != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_RESTRICT, NULL);

  relation->to_restrict = to_restrict;
  g_object_ref (to_restrict);

  relation->func = predicate_func;
  relation->func_data = user_data;

  init_index (relation);

  g_signal_connect_swapped (to_restrict, "insert-tuple",
                            G_CALLBACK (h_relation_restrict_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_restrict, "delete-tuple",
                            G_CALLBACK (h_relation_restrict_delete_tuple),
                            relation);

  return H_RELATION (relation);
}

typedef struct
{
  HTuple *tuple;
  GSList *list;
}
TupleHasParams;

static gboolean
tuple_has (HTuple         *tuple,
           TupleHasParams *params)
{
  return h_tuple_equal_elems_l (tuple, params->tuple, params->list);
}

static TupleHasParams *
tuple_has_params_new (HRelation *relation,
                      GSList    *list)
{
  TupleHasParams *params;
  GSList         *iter;
  const gchar    *name;

  params = g_new (TupleHasParams, 1);

  params->tuple = h_tuple_new ();
  params->list  = NULL;

  for (iter = list; iter; iter = iter->next->next)
    {
      name = (const gchar *) list->data;

      h_tuple_insert (params->tuple, name,
                      h_relation_get_attr (relation, name)->type,
                      list->next->data);

      params->list = g_slist_append (params->list, g_strdup (name));
    }

  return params;
}

static void
tuple_has_params_free (TupleHasParams *params)
{
  h_string_list_free (params->list);
  h_tuple_free (params->tuple);
  g_free (params);
}

/**
 * h_relation_restrict_attrs:
 * @relation: relation to restrict
 * @...: name and value of an attribute, followed by another and so
 *   on, then a %NULL.
 *
 * Restricts @relation to the set of tuples which contain the
 * attribute values listed.
 *
 * Return value: restricted relation
 **/
HRelation *
h_relation_restrict_attrs (HRelation *to_restrict,
                           ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  va_start (args, to_restrict);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_restrict_attrs_l (to_restrict, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_restrict_attrs_l:
 * @relation: relation to restrict
 * @list: list of attribute names and their values
 *
 * A variation of h_relation_restrict_attrs(), taking a #GSList for
 * the attribute list.
 *
 * Return value: restricted relation
 **/
HRelation *
h_relation_restrict_attrs_l (HRelation *to_restrict,
                             GSList    *list)
{
  HRelation      *relation;
  TupleHasParams *params;

  params = tuple_has_params_new (to_restrict, list);

  relation = h_relation_restrict (to_restrict, (HTuplePredFunc) tuple_has,
                                  params);

  g_object_set_data_full (G_OBJECT (relation), "tuple-has-params", params,
                          (GDestroyNotify) tuple_has_params_free);

  return relation;
}
