/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_sort.h"
#include "misc.h"
#include <stdio.h>

/**
 * HRelationSort:
 *
 * A #HRelation representing a "sorted relation".
 **/
G_DEFINE_TYPE (HRelationSort, h_relation_sort, H_TYPE_RELATION)

static void  h_relation_sort_dispose        (GObject       *object);
static void  h_relation_sort_finalize       (GObject       *object);
static void  h_relation_sort_print          (HRelation     *relation);
static HAttribute *h_relation_sort_get_attr (HRelation     *relation,
                                             const gchar   *name);
static guint h_relation_sort_get_degree     (HRelation     *relation);
static guint h_relation_sort_get_size       (HRelation     *relation);
static void  h_relation_sort_for_each_attr  (HRelation     *relation,
                                             HAttrFunc      func,
                                             gpointer       user_data);
static void  h_relation_sort_for_each_tuple (HRelation     *relation,
                                             HTupleFunc     func,
                                             gpointer       user_data);
static void  h_relation_sort_insert_tuple   (HRelationSort *self,
                                             HTuple        *tuple);
static void  h_relation_sort_delete_tuple   (HRelationSort *self,
                                             HTuple        *tuple);

static void
h_relation_sort_class_init (HRelationSortClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  object_class->dispose  = h_relation_sort_dispose;
  object_class->finalize = h_relation_sort_finalize;

  relation_class->print = h_relation_sort_print;
  relation_class->get_attr = h_relation_sort_get_attr;
  relation_class->get_degree = h_relation_sort_get_degree;
  relation_class->get_size = h_relation_sort_get_size;
  relation_class->for_each_attr = h_relation_sort_for_each_attr;
  relation_class->for_each_tuple = h_relation_sort_for_each_tuple;
}

static void
h_relation_sort_init (HRelationSort *self)
{
  self->to_sort = NULL;
  self->compare_func = NULL;
  self->index = NULL;
}

static void
h_relation_sort_dispose (GObject *object)
{
  G_OBJECT_CLASS (h_relation_sort_parent_class)->dispose (object);
}

static void
h_relation_sort_finalize (GObject *object)
{
  HRelationSort *self = H_RELATION_SORT (object);

  g_signal_handlers_disconnect_by_func (self->to_sort,
                                        (gpointer) h_relation_sort_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (self->to_sort,
                                        (gpointer) h_relation_sort_delete_tuple,
                                        self);

  g_object_unref (self->to_sort);

  g_tree_destroy (self->index);

  G_OBJECT_CLASS (h_relation_sort_parent_class)->finalize (object);
}

static void
h_relation_sort_print (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);
  gchar *str;

  g_return_if_fail (self != NULL);

  printf ("Relation to sort:\n");

  str = h_relation_get_name (self->to_sort);
  printf ("  Name: %s\n", str ? str : "(none)");
  g_free (str);

  str = h_relation_get_desc (self->to_sort);
  printf ("  Description: %s\n", str ? str : "(none)");
  g_free (str);
}

static HAttribute *
h_relation_sort_get_attr (HRelation   *relation,
                          const gchar *name)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_val_if_fail (self != NULL, NULL);

  return h_relation_get_attr (self->to_sort, name);
}

static void
h_relation_sort_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_if_fail (self != NULL);

  h_relation_for_each_attr (self->to_sort, func, user_data);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
IndexIterateParams;

static void
index_iterate (HTuple             *key,
               GList              *list,
               IndexIterateParams *params)
{
  GList *iter;

  for (iter = list; iter != NULL; iter = iter->next)
    {
      params->func ((HTuple *) iter->data, params->data);
    }
}

static void
h_relation_sort_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationSort *self = H_RELATION_SORT (relation);
  IndexIterateParams params;

  g_return_if_fail (self != NULL);

  params.func = func;
  params.data = user_data;

  g_tree_foreach (self->index, (GTraverseFunc) index_iterate, &params);
}


static guint
h_relation_sort_get_size (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_size (self->to_sort);
}

static guint
h_relation_sort_get_degree (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_degree (self->to_sort);
}

static GList *
list_find_custom_with_data (GList            *list,
                            gconstpointer     data,
                            GCompareDataFunc  func,
                            gpointer          func_data)
{
  GList *iter;

  for (iter = list; iter; iter = iter->next)
    {
      if ((*func) (data, iter->data, func_data) == 0)
        return iter;
    }

  return NULL;
}

static void
index_insert (HTuple        *tuple,
              HRelationSort *relation)
{
  GList *list;

  list = g_tree_lookup (relation->index, tuple);

  if (list)
    g_tree_steal (relation->index, tuple);

  list = g_list_append (list, tuple);
  g_tree_insert (relation->index, tuple, list);
}

static void
index_delete (HTuple        *tuple,
              HRelationSort *relation)
{
  GList *list, *node;

  list = g_tree_lookup (relation->index, tuple);
  if (!list)
    return;

  node = list_find_custom_with_data (list, tuple,
                                     relation->compare_func,
                                     relation->compare_func_data);
  if (!node)
    return;

  list = g_list_delete_link (list, node);

  g_tree_steal (relation->index, tuple);

  if (list)
    g_tree_insert (relation->index, tuple, list);
}

static void
h_relation_sort_insert_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  index_insert (tuple, self);

  g_signal_emit_by_name (self, "insert-tuple", tuple);
}

static void
h_relation_sort_delete_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  index_delete (tuple, self);

  g_signal_emit_by_name (self, "delete-tuple", tuple);
}

static void
build_index (HRelationSort *self)
{
  self->index = g_tree_new_full (self->compare_func,
                                 self->compare_func_data,
                                 NULL,
                                 (GDestroyNotify) g_list_free);

  h_relation_for_each_tuple (self->to_sort,
                             (HTupleFunc) index_insert,
                             self);
}

/**
 * h_relation_sort:
 * @relation: a relation to sort
 * @compare_func: function to compare a pair of tuples in @relation
 * @user_data: Custom data passed to @compare_func
 *
 * Creates a sorted relation.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort (HRelation        *to_sort,
                 GCompareDataFunc  compare_func,
                 gpointer          user_data)
{
  HRelationSort *relation;

  g_return_val_if_fail (to_sort != NULL, NULL);
  g_return_val_if_fail (compare_func != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_SORT, NULL);

  relation->to_sort = to_sort;
  g_object_ref (to_sort);

  relation->compare_func = compare_func;
  relation->compare_func_data = user_data;

  g_signal_connect_swapped (to_sort, "insert-tuple",
                            G_CALLBACK (h_relation_sort_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_sort, "delete-tuple",
                            G_CALLBACK (h_relation_sort_delete_tuple),
                            relation);

  build_index (relation);

  return H_RELATION (relation);
}

/**
 * h_relation_sort_by_attrs:
 * @relation: relation to sort
 * @...: name of first attribute, followed by the second and so on,
 *   then a %NULL
 *
 * Creates a relation sorted by attributes. Each pair of tuples are
 * compared attribute by attribute in the order listed.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort_by_attrs (HRelation *to_sort,
                          ...)
{
  va_list    args;
  GSList    *list;
  HRelation *relation;

  va_start (args, to_sort);
  list = h_va_list_to_slist (args);
  va_end (args);

  relation = h_relation_sort_by_attrs_l (to_sort, list);
  g_slist_free (list);

  return relation;
}

/**
 * h_relation_sort_by_attrs_l:
 * @relation: relation to sort
 * @list: list of attribute names
 *
 * A variation of h_relation_sort_by_attrs(), taking a #GSList for the
 * attribute name list.
 *
 * Return value: sorted relation
 **/
HRelation *
h_relation_sort_by_attrs_l (HRelation *to_sort,
                            GSList    *list)
{
  HRelation *relation;
  GSList    *list_copy;

  list_copy = h_string_list_copy (list);

  relation = h_relation_sort (to_sort,
                              (GCompareDataFunc) h_tuple_compare_elems_l,
                              list_copy);

  g_object_set_data_full (G_OBJECT (relation), "sort-attrs", list_copy,
                          (GDestroyNotify) h_string_list_free);

  return relation;
}
