/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <glib-object.h>

#include <bmpx/plugin.h>
#include <bmpx/playlist.h>
#include <bmpx/metadata.h>
#include <bmpx/list.h>
#include <bmpx/plugin_interfaces.h>

#include "libflow_shuffle.h"

static char shuffle_history_data[] = "dummy data";
static GHashTable *shuffle_history = NULL;

static void
row_guid_free_key_func (RowGUID *key)
{
    g_free (key);
}

static RowGUID*
get_fileitem_guid (BmpPlaylistItem *playlist, gint track)
{
    GList *list_item;
    BmpFileItem *file_item;

    list_item = g_list_nth (playlist->list->items, track);
    if (!list_item)
      return NULL;
    file_item = list_item->data;
    if (!file_item)
      return NULL;
    
    return file_item->guid;
}

static void
flow_shuffle_history_free ()
{
    if (shuffle_history)
      {
	g_hash_table_destroy (shuffle_history);
	shuffle_history = NULL;
      }
}

static void
flow_shuffle_history_item_insert (RowGUID *track)
{
    RowGUID *temp;

    temp = g_malloc0 (sizeof (RowGUID));
    memcpy (temp, track, sizeof (RowGUID));
	
    g_hash_table_insert (shuffle_history, temp, shuffle_history_data);
}

static void
flow_shuffle_history_init (BmpPlaylistItem *playlist,
			   gint list_idx,
			   RowGUID *current_guid)
{
    if (!shuffle_history)
      {
	shuffle_history = g_hash_table_new_full ((GHashFunc) row_guid_hash_func,
					    	 (GEqualFunc) row_guid_equal_func,
						 (GDestroyNotify) row_guid_free_key_func,
						 NULL);

	/* Append current track, so that we do not play it twice */
	flow_shuffle_history_item_insert (current_guid);
      }
}

static gboolean
flow_shuffle_history_item_played (RowGUID *track)
{
    if (g_hash_table_lookup (shuffle_history, track))
      return TRUE;

    return FALSE;
}

static gint
flow_shuffle_history_item_next (BmpPlaylistItem *playlist,
			       gint list_idx,
			       gint list_length,
			       RowGUID *current_guid)
{
    GRand *rand;
    RowGUID *guid;
    gint track, i = 0;

    rand = g_rand_new ();
   
    do
      {
	/* FIXME: g_rand_int_range () might skip up to 1% of the range
	 *        within the first list_length number of calls */
	if (i++ >= list_length)
	  {
	    g_rand_free (rand);
	    return -1;
	  }
	
	track = g_rand_int_range (rand, 0, list_length);
	guid = get_fileitem_guid (playlist, track);
	if (!guid)
	  continue;
      }
    while (row_guid_equal_func (guid, current_guid) ||
	   flow_shuffle_history_item_played (guid));

    g_rand_free (rand);

    flow_shuffle_history_item_insert (guid);

    return track;
}

/* Shuffle Flow Control */
static BmpPlaybackHistoryItem*
flow_shuffle_item_next (BmpPluginFlowInterface *self, const BmpListType *list, gint _idx)
{
    BmpPlaylistItem *playlist;
    BmpFileItem *file_item;
    BmpPlaybackHistoryItem *history_item;
    RowGUID *current_guid;
    gint length;
    gint current;
    gint track;

    playlist = BMP_PLAYLIST_ITEM(g_list_nth_data(list->items, _idx)); 

    length = g_list_length (playlist->list->items);

    current = playlist->list->current;

    current_guid = get_fileitem_guid (playlist, current);
    if (!current_guid)
      return NULL;

    flow_shuffle_history_init (playlist, _idx, current_guid);

    track = flow_shuffle_history_item_next (playlist, _idx, length,
					    current_guid);
    if (track == -1)
      return NULL;

    history_item = g_new0 (BmpPlaybackHistoryItem,1);

    history_item->list_time = playlist->list->time;

    history_item->list_name = g_strdup (playlist->listname);

    file_item = BMP_FILE_ITEM(g_list_nth_data(playlist->list->items, track));
    history_item->guid = g_malloc0(sizeof(RowGUID));
    memcpy (history_item->guid, file_item->guid, sizeof(RowGUID));

    return history_item;
}

static BmpPlaybackHistoryItem*
flow_shuffle_item_prev  (BmpPluginFlowInterface *self, const BmpListType *list, gint _idx)
{
    BmpPlaylistItem *playlist;
    BmpFileItem *file_item;
    BmpPlaybackHistoryItem *history_item;
    RowGUID *current_guid;
    gint length;
    gint current;
    gint track;

    playlist = BMP_PLAYLIST_ITEM(g_list_nth_data(list->items, _idx)); 

    length = g_list_length (playlist->list->items)-1;

    current = playlist->list->current;

    current_guid = get_fileitem_guid (playlist, current);
    if (!current_guid)
      return NULL;

    flow_shuffle_history_init (playlist, _idx, current_guid);

    track = flow_shuffle_history_item_next (playlist, _idx, length,
					    current_guid);
    if (track == -1)
      return NULL;

    history_item = g_new0 (BmpPlaybackHistoryItem,1);

    history_item->list_time = playlist->list->time;

    history_item->list_name = g_strdup (playlist->listname);

    file_item = BMP_FILE_ITEM(g_list_nth_data(playlist->list->items, track));
    history_item->guid = g_malloc0(sizeof(RowGUID));
    memcpy (history_item->guid, file_item->guid, sizeof(RowGUID));

    return history_item;
}

static void
reset_state (BmpPluginFlowInterface *self)
{
    flow_shuffle_history_free ();
}

/* IFaces */
static gchar* _metadata[] = {
    "Shuffle",
    "Milosz Derezynski and Martin Schlemmer",
    "(C) GPL 2005",
    "http://bmpx.berlios.de",
    "internalerror@gmail.com",
};

static const gchar**
get_metadata (BmpPlugMetadataInterface *self)
{
    return (const gchar**)_metadata;
}

static void
metadata_init (gpointer         g_iface,
	       gpointer         iface_data)
{
	BmpPlugMetadataInterface *klass = (BmpPlugMetadataInterface *)g_iface;

	klass->get_metadata = get_metadata;
}


void
flow_shuffle_interface_init (gpointer         g_iface)
{
        BmpPluginFlowInterface *klass = (BmpPluginFlowInterface *)g_iface;

        klass->item_next   = flow_shuffle_item_next;
        klass->item_prev   = flow_shuffle_item_prev;
        klass->reset_state = reset_state; 
}


GType plugin_flow_shuffle_type = 0;

void 
plugin_flow_shuffle_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginFlowShuffleClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginFlowShuffle),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                static const GInterfaceInfo interface_flow_shuffle_info = {
                        (GInterfaceInitFunc) flow_shuffle_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                static const GInterfaceInfo interface_metadata_info = {
                        (GInterfaceInitFunc) metadata_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                plugin_flow_shuffle_type = g_type_module_register_type (module,
					   G_TYPE_OBJECT,
					  "BmpPluginFlowShuffleType",
					   &object_info, 0);

                g_type_module_add_interface (module,
					     plugin_flow_shuffle_type,
                                             BMP_TYPE_PLUG_METADATA,
                                             &interface_metadata_info);

                g_type_module_add_interface (module,
					     plugin_flow_shuffle_type,
                                             BMP_TYPE_PLUGIN_FLOW,
                                             &interface_flow_shuffle_info);
}
