#include <config.h>

#include <stdio.h>
#ifndef __USE_BSD
# define __USE_BSD	/* Needed for mkdtemp () */
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#ifdef ENABLE_NLS
# include <libintl.h>
#endif

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include <bmpx/util.h>
#include <bmpx/uri.h>
#include <bmpx/vfs.h>
#include <bmpx/plugin.h>
#include <bmpx/plugin_interfaces.h>

#include "libtransport_http.h"

#ifdef HAVE_FTS_H
#  include <fts.h>
#endif

static VFSHandle *handle = NULL;
static GCond	 *handle_read_cond = NULL;

/* CURL Auxilliary stuff */
struct MemoryStruct {
    char *memory;
    size_t size;
};
static void *myrealloc (void *ptr, size_t size)
{
    /* There might be a realloc () out there that doesn't like reallocing
       NULL pointers, so we take care of it here */
    if (ptr)
	return realloc (ptr, size);
    else
	return malloc (size);
}
static size_t
WriteMemoryCallback (void *ptr, size_t size, size_t nmemb, void *data)
{
    size_t realsize = size * nmemb;
    struct MemoryStruct *mem = (struct MemoryStruct *)data;

    mem->memory = (char *)myrealloc (mem->memory, mem->size + realsize + 1);
    if (mem->memory)
	{
	    memcpy (& (mem->memory[mem->size]), ptr, realsize);
	    mem->size += realsize;
	    mem->memory[mem->size] = 0;
	}

    return realsize;
}
/* END CURL stuff */

gboolean
transport_http_can_process (BmpPluginTransportInterface *self, const gchar *uri)
{
    if (!g_ascii_strncasecmp("http",uri,4))
	return TRUE;

    return FALSE;
}

gboolean
transport_http_handle_set (BmpPluginTransportInterface *self, VFSHandle *_handle)
{
    if (_handle == NULL) 
	return FALSE;
    
    if (!_handle->valid)
	return FALSE;

    handle = _handle;

    return TRUE;
}

VFSHandle*
transport_http_handle_get (BmpPluginTransportInterface *self)
{
    return handle;
}

gboolean
transport_http_handle_remove (BmpPluginTransportInterface *self)
{
    handle = NULL;

    return TRUE;
}

struct fetch_thread_data_t {
    GMutex   *mutex;
    CURL     *handle;
    gboolean  done;
};
#define FETCH_THREAD_DATA(_data) ((struct fetch_thread_data_t *)(_data))

static gpointer
curl_fetch_thread (gpointer data)
{
    /* get it! */
    curl_easy_perform (FETCH_THREAD_DATA (data)->handle);

    FETCH_THREAD_DATA(data)->done = TRUE;

    /* we are done, so transport_http_handle_read() can return the data */
    g_mutex_lock (FETCH_THREAD_DATA (data)->mutex);
    g_cond_signal (handle_read_cond);
    g_mutex_unlock (FETCH_THREAD_DATA (data)->mutex);

    g_thread_exit (NULL);

    /* shutup gcc */
    return NULL;
}

gboolean
transport_http_handle_read (BmpPluginTransportInterface *self)
{
    GThread *fetch_thread = NULL;
    struct fetch_thread_data_t thread_data;
    CURL *curl_handle;
    GTimeVal _tv;

    if (g_strrstr(handle->uri+6,":") && (!is_audio_file(handle->uri))) // FIXME: Too lazy check for direct socket address URL
      {
	handle->buffer = NULL;
	handle->uri_list = g_list_append (handle->uri_list, g_strdup(handle->uri));
	handle->need_container = FALSE;
	return TRUE;
      }

    struct MemoryStruct chunk;

    chunk.memory = NULL; /* we expect realloc (NULL, size) to work */
    chunk.size = 0;    /* no data at this point */

    curl_global_init (CURL_GLOBAL_ALL);

    /* init the curl session */
    curl_handle = curl_easy_init ();

    /* specify URL to get */
    curl_easy_setopt (curl_handle, CURLOPT_URL, handle->uri);

    /* send all data to this function  */
    curl_easy_setopt (curl_handle, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);

    /* we pass our 'chunk' struct to the callback function */
    curl_easy_setopt (curl_handle, CURLOPT_WRITEDATA, (void *)&chunk);

    /* some servers don't like requests that are made without a user-agent
       field, so we provide one */
    curl_easy_setopt (curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

    /* get a new mutex, and lock it - the thread will unlock it when done */
    thread_data.handle = curl_handle;
    thread_data.done  = FALSE;

    handle_read_cond = g_cond_new ();

    thread_data.mutex = g_mutex_new ();
    g_mutex_lock (thread_data.mutex);

    /* setup data to be passed to thread, and create new thread to handle
     * fetching the http data */
    fetch_thread = g_thread_create (curl_fetch_thread,
				   &thread_data,
				   FALSE,
				   NULL);

    /* wait while we busy fetching the data and make sure signals are passed */
    while (42)
      {
	while (gtk_events_pending ())
	  gtk_main_iteration ();

	g_get_current_time (&_tv);  
	g_time_val_add (&_tv, 10000);

        if (g_cond_timed_wait(handle_read_cond, thread_data.mutex, &_tv))
          break;
      }

    /* all done, unlock and free mutex */
    g_mutex_unlock (thread_data.mutex);
    g_mutex_free (thread_data.mutex);

    /* cleanup curl stuff */
    curl_easy_cleanup (curl_handle);

    /*
     * Now, our chunk.memory points to a memory block that is chunk.size
     * bytes big and contains the remote file.
     *
     * Do something nice with it!
     *
     * You should be aware of the fact that at this point we might have an
     * allocated data block, and nothing has yet deallocated that data. So when
     * you're done with it, you should free () it as a nice application.
     */

    handle->buffer = chunk.memory;

    return TRUE;
}

static void
transport_http_interface_init (gpointer         g_iface,
                               gpointer         iface_data)
{
        BmpPluginTransportInterface *klass = (BmpPluginTransportInterface *)g_iface;

	klass->can_process    = transport_http_can_process;
	klass->handle_set     = transport_http_handle_set;
	klass->handle_get     = transport_http_handle_get;
	klass->handle_remove  = transport_http_handle_remove;
	klass->handle_read    = transport_http_handle_read;
}

GType plugin_transport_http_type = 0;

void 
plugin_transport_http_register_type (GTypeModule *module)
{
	    static const GTypeInfo object_info = {
                        sizeof (BmpPluginTransportHttpClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (BmpPluginTransportHttp),
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };

                static const GInterfaceInfo interface_transport_http_info = {
                        (GInterfaceInitFunc) transport_http_interface_init,    /* interface_init */
                        NULL,                                       /* interface_finalize */
                        NULL                                        /* interface_data */
                };

                plugin_transport_http_type = g_type_module_register_type (module,
					     G_TYPE_OBJECT,
					     "BmpPluginTransportHttpType",
					     &object_info, 0);

                g_type_module_add_interface (module,
					     plugin_transport_http_type,
                                             BMP_TYPE_PLUGIN_TRANSPORT,
                                             &interface_transport_http_info);
}
