#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <config.h>

#include <glib.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#endif

#include <bmpx/metadata.h>

#include <libhrel/types.h>
#include <libhrel/tuple.h>

/* Magic version key for LCMFS.  It should always be 10 chars long */
const char lcmfs_version_magic[] = "LCMFS 0.02";
/* Magic key for each LCMFS record.  It should always be 10 chars long */
const char lcmfs_record_magic[] = "LCMFS Rcrd";
/* Length of magic version key.  It is the 10 chars of the version
 * key plus the terminating 0 */
const gint lcmfs_magic_length = 11;

typedef struct _LCMFS LCMFS;

struct _LCMFS {
    int		   i_size;
    GHashTable	  *i;

    gchar	  *index;
    gchar	  *records;

    FILE	  *f_index; 
    FILE	  *f_records; 
};

static LCMFS *lcmfs;

void
lcmfs_init ()
{
  lcmfs = g_new0 (LCMFS,1);
  
  lcmfs->i_size = 
      0;
  lcmfs->i = 
      g_hash_table_new (g_str_hash, g_str_equal);
}

void
lcmfs_open (const gchar *lcmfs_base)
{
  gint retval = 0;
  
  lcmfs->index =
    g_strconcat (lcmfs_base, ".lif", NULL); // LCMFS Index  File
  lcmfs->records = 
    g_strconcat (lcmfs_base, ".lrd", NULL); // LCMFS Record Data

  // Open Index
  lcmfs->f_index = fopen (lcmfs->index, "r");
  if (lcmfs->f_index)
    {
      // Read index
      char lcmfs_magic_str[lcmfs_magic_length];
      gint n, count;

      count = fread (lcmfs_magic_str, sizeof(char), lcmfs_magic_length,
	             lcmfs->f_index);
      lcmfs_magic_str[lcmfs_magic_length - 1] = '\0';
      if (count != lcmfs_magic_length || strcmp(lcmfs_magic_str, lcmfs_version_magic) != 0)
	{
	  /* Index reading failed or wrong version. */
	  g_message ("LCMFS Index version mismatch.");
	  goto done;
	}
      
      fread (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
      for (n = 0; n < lcmfs->i_size; n++)
	{
	  uint16_t key_size;
	  gint     offset;
	  gchar   *pkey;

	  count = fread (&key_size, sizeof(uint16_t), 1, lcmfs->f_index);
	  if (count != 1)
	    goto read_error;
	  
	  pkey = g_malloc0 (key_size+1);
	  count = fread (pkey, sizeof(char), key_size, lcmfs->f_index);
	  if (count != key_size)
	    goto read_error;
	  
	  count = fread (&offset, sizeof(gint), 1, lcmfs->f_index);
	  if (count != 1)
	    goto read_error;

	  g_hash_table_insert (lcmfs->i, pkey, GINT_TO_POINTER (offset)); 

	  continue;

read_error:
	  g_message ("Error reading LCMFS Index.");
	  goto done;
	}

      retval = 1;
      
done:
      fclose (lcmfs->f_index);
    } 

  g_message ("%s: i_size: '%d'", G_STRLOC, lcmfs->i_size);

  if (retval)
    // Open Records
    lcmfs->f_records = fopen (lcmfs->records, "a+");
  else
    // Reading of index failed, truncate records file
    lcmfs->f_records = fopen (lcmfs->records, "w+");
}

gint
lcmfs_get_index_size (void)
{
  return lcmfs->i_size;
}

gboolean
lcmfs_record_exists (const gchar *pkey)
{
  return g_hash_table_lookup (lcmfs->i, pkey) ? TRUE : FALSE;
}

void
record_read_mem (HTuple *record, unsigned char **data)
{
  gint n;

  *data += lcmfs_magic_length;

  for (n = 0; n < BMP_DATUM_NONE ; n++)
    {
      gint *metadatum_id_ptr;
      gint metadatum_id;
      
      metadatum_id_ptr = (gint *)(*data);
      metadatum_id = *metadatum_id_ptr;
      *data += sizeof(gint);
 
      if (bmp_metadata_get_type (metadatum_id) == G_TYPE_INT)
        {
          gint *value;
 
          value = (gint*)(*data);
 
	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (metadatum_id),
			  G_TYPE_INT,
			  GINT_TO_POINTER(*value));
 
          *data += sizeof(gint);
        }
      else if (bmp_metadata_get_type (metadatum_id) == G_TYPE_STRING)
        {
          char *value;
 
          value = (char *)(*data);
  
	  if ((!value) || (strlen(value)==0))
	    value = "";
 
	  h_tuple_insert (record,
			  bmp_metadata_get_id_static (metadatum_id),
			  G_TYPE_STRING,
			  value); 
 
           /* string length + terminating 0 */
          *data += strlen(value) + 1;
        }
    }
}

void
lcmfs_read (BmpMetadataCache *cache)
{
  unsigned char *current_ptr, *current_ptr_save;
  FILE		*fp;
  gint		 n, ret, fd;

  struct stat	 st;

  ret = stat (lcmfs->records, &st);
  if (ret == -1)
    {
      /* File does not exits? */
      g_message ("LCMFS Records do not exist.");
      return;
    }

  fp = fopen (lcmfs->records, "r");

  if (!fp)
    {
      g_message ("Could not open LCMFS Records.");
      return;
    }

  fd = fileno (fp);

  current_ptr = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
  current_ptr_save = current_ptr;
  fclose(fp);

  if (!current_ptr)
    {
      g_message ("Could not mmap LCMFS Records.");
      return;
    }

  for (n = 0; n < lcmfs->i_size; n++)
    {
      HTuple *record;
      char lcmfs_magic_str[lcmfs_magic_length];

      snprintf(lcmfs_magic_str, lcmfs_magic_length, "%s", current_ptr);
      if (strcmp(lcmfs_magic_str, lcmfs_record_magic) != 0)
	{
	  g_message ("LCMFS Record %i corrupted.", n + 1);
	  continue;
	}

      record = bmp_metadata_tuple_create_default ();
      record_read_mem (record, &current_ptr);

      _bmp_metadata_cache_tuple_inserted (cache, record);

#ifdef HAVE_GUI
      while (gtk_events_pending ())
	gtk_main_iteration ();
#endif
    }

  munmap(current_ptr_save, st.st_size);

  if (st.st_size > (size_t)(current_ptr - current_ptr_save))
    {
      /* There are more data in the file than there should be,
       * so truncate it to the correct size. */
      fclose (lcmfs->f_records);
      fp = fopen (lcmfs->records, "a+");
      if (!fp)
	return;
      fd = fileno (fp);
      ftruncate (fd, current_ptr - current_ptr_save);
      fclose (fp);
      lcmfs->f_records = fopen (lcmfs->records, "a+");
    }
}

gboolean
lcmfs_record_add (HTuple *record)
{
  HTupleElem *elem_pkey, *elem;
  gint	     new_offset, count, n = 0;

  // Just for the case we check if the record exists
  elem_pkey = h_tuple_get (record, (gpointer)bmp_metadata_get_id_static (BMP_METADATA_PKEY));
  if (lcmfs_record_exists (elem_pkey->value))
      return FALSE;

  // Seek to the end of the records file
  fseek (lcmfs->f_records, 0L, SEEK_END);
  new_offset = ftell (lcmfs->f_records);

  count = fwrite (lcmfs_record_magic, lcmfs_magic_length, 1,
		  lcmfs->f_records);
  if (count != 1)
    goto error;

  for (n = 0; n < BMP_DATUM_NONE; n++)
    {
      elem = h_tuple_get (record, bmp_metadata_get_id_static (n)); 

      if (elem->type == G_TYPE_INT)
	{
	  gint v;

	  v = GPOINTER_TO_INT (elem->value);

	  count = fwrite (&n, sizeof(gint), 1, lcmfs->f_records);
	  if (count != 1)
	    goto error;
          count = fwrite (&v, sizeof(gint), 1, lcmfs->f_records);
	  if (count != 1)
	    goto error;
	}
      else if (elem->type == G_TYPE_STRING)
	{
	  gint length = strlen(elem->value)+1;
	  
	  count = fwrite (&n, sizeof(gint), 1, lcmfs->f_records);
	  if (count != 1)
	    goto error;
          count = fwrite (elem->value, sizeof(char), length, lcmfs->f_records);
	  if (count != length)
	    goto error;
	}
    }

  g_hash_table_insert (lcmfs->i, g_strdup (elem_pkey->value), GINT_TO_POINTER (new_offset));
  lcmfs->i_size++;

  return TRUE;

error:
  g_message("Failed to add LCMFS Record");
  return FALSE;
}

static void
_write_index_record (gpointer key, gpointer value, gpointer data)
{
  uint16_t  key_size;
  gchar    *pkey   = (gchar*)key;
  gint      offset = GPOINTER_TO_INT (value);

  key_size = strlen (pkey);

  // FIXME: Any way to handle errors here ?
  fwrite (&key_size, sizeof(uint16_t), 1, lcmfs->f_index); 
  fwrite (pkey, sizeof(char), key_size, lcmfs->f_index);
  fwrite (&offset, sizeof(gint), 1, lcmfs->f_index);
}

void
lcmfs_close ()
{
  gint count;
  
  lcmfs->f_index = fopen (lcmfs->index, "w");
  if (lcmfs->f_index)
    {
      // Write index
      count = fwrite (lcmfs_version_magic, sizeof(char), lcmfs_magic_length,
	              lcmfs->f_index);
      if (count != lcmfs_magic_length)
	g_message ("Could not write LCMFS index magic key");
      count = fwrite (&lcmfs->i_size, sizeof(int), 1, lcmfs->f_index);
      if (count != 1)
	g_message ("Could not write LCMFS index size");
      g_hash_table_foreach (lcmfs->i, (GHFunc)_write_index_record, NULL);
      fclose (lcmfs->f_index);
    } 
  else
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s: Unable to write index file", G_STRLOC);
      // ERROR
    }

  fclose (lcmfs->f_records);
}

