#include <config.h>

#include <glib.h>
#include <gmodule.h>
#include <stdlib.h>
#include <glib/gi18n.h>

#include "main.h"
#include "gc.h"

#include "bmpx/util.h"
#include "bmpx/plugin.h"
#include "bmpx/plugin_interfaces.h"

#include "plugin_priv.h"

GHashTable   *plugins_table[BMP_PLUGIN_N_TYPES];

#ifdef HAVE_GUI
GtkTreeStore *plugins_about = NULL;

static GtkTreeIter iter;
static GtkTreeIter iter_parent;
static GtkTreeIter iter_child;
#endif

enum {
    LIB_BASENAME,
    LIB_PLUGNAME,
    LIB_SUFFIX
};

#define IS_MODULE(basename) (str_has_suffix_nocase(basename, G_MODULE_SUFFIX))

typedef struct {
    gchar *name;
    gint   table_type;
} ForeachTuple;

gboolean load_plugins (const gchar *path,
		       const gchar *basename,
		       gpointer     user_data)
{
    GObject   *plugin_instance;
    BmpPlugin *plugin;

    ForeachTuple *tuple = (ForeachTuple*)user_data;

    gchar **subs, **_save;
    gchar  *mpath, *dir, *name;

    const gchar **metadata = NULL;
    GType  *iface_ids;
    guint n_ifaces;
    gint n;

    NGC *gc;

    if (!IS_MODULE(basename))
	return FALSE;

    gc = gc_new();
    gc_mark (gc);

    g_message ("%s: %s plugin: %s", G_STRLOC, tuple->name, basename);
    _save = subs = gc_collect_d (gc, g_strsplit_set (basename, "_.", -1), (GDestroyNotify)g_strfreev);

    dir =
	gc_collect (gc, g_build_filename (PLUGIN_DIR,tuple->name,NULL));
    name =
	gc_collect (gc, g_strconcat (tuple->name,"_",subs[LIB_PLUGNAME],NULL));
    mpath = 
	gc_collect (gc, g_module_build_path (dir, name));

    plugin =
        bmp_plugin_new (mpath);	
    g_type_module_use (G_TYPE_MODULE(plugin));

    if (!bmp_plugin_valid (plugin))
      {
	gc_free (gc);
	gc_destroy (gc);
	return FALSE;
      }

    plugin_instance =
	bmp_plugin_create_instance (plugin);

#ifdef HAVE_GUI
    gtk_tree_store_append (plugins_about, &iter, &iter_parent);
    gtk_tree_store_set (plugins_about, &iter, 1, basename, -1);

    if (BMP_IS_PLUG_METADATA(plugin_instance))
      {

	metadata = bmp_plug_metadata_get (BMP_PLUG_METADATA(plugin_instance));
	if (metadata)
	  {
	    for (n=0;n<METADATA_N_FIELDS;n++)
		gtk_tree_store_set (plugins_about, &iter, n+2, metadata[n], -1);
	  }
      }

    gtk_tree_store_set (plugins_about, &iter, METADATA_N_FIELDS+2, G_OBJECT_TYPE_NAME(plugin_instance), -1);
    iface_ids = g_type_interfaces (G_OBJECT_TYPE(plugin_instance), &n_ifaces);

    for (n=0;n<n_ifaces;n++)
      {
	gtk_tree_store_append (plugins_about, &iter_child, &iter);
	gtk_tree_store_set (plugins_about, &iter_child, METADATA_N_FIELDS+3, g_type_name(iface_ids[n]), -1);
      }
    g_free(iface_ids);
#endif


    // FIMXE: Not sure this is the right way to go...
    g_hash_table_insert (plugins_table[tuple->table_type], g_strdup(subs[LIB_PLUGNAME]), plugin_instance);

    gc_free (gc); 
    gc_destroy (gc);
    return FALSE;
}


static gchar* plugin_type_names[] = {
    "flow",
    "container",
    "transport",
    NULL,
};

static gchar* plugin_col_names[] = {
    "Flow",
    "Container",
    "Transport",
    NULL,
};


void
plugins_initialize (void)
{
    GError *error = NULL;
    gint n;

#ifdef HAVE_GUI
    plugins_about = gtk_tree_store_new (9, 
					G_TYPE_STRING, 
					G_TYPE_STRING, 
					G_TYPE_STRING, 
					G_TYPE_STRING, 
					G_TYPE_STRING, 
					G_TYPE_STRING, 
					G_TYPE_STRING,
					G_TYPE_STRING,
					G_TYPE_STRING);
#endif
    
    for (n = 0; n < BMP_PLUGIN_N_TYPES; n++)
      {
	 plugins_table[n] = 
	    g_hash_table_new (g_str_hash, g_str_equal);
      }
   
    for (n = 0; plugin_type_names[n]; n++)
     {
	ForeachTuple *tuple;
	gchar *path;

	path = 
	    g_build_filename (PLUGIN_DIR, plugin_type_names[n], NULL);
	tuple =
	    g_new0 (ForeachTuple,1);
	
	tuple->name = plugin_type_names[n];
	tuple->table_type = n;

#ifdef HAVE_GUI
	gtk_tree_store_append (plugins_about, &iter_parent, NULL); 
	gtk_tree_store_set (plugins_about, &iter_parent, 0, plugin_col_names[n], -1);
#endif
	
	dir_foreach (path, load_plugins, tuple, &error);

	if (error)
	  {
	    g_message ("%s: %s", G_STRLOC, error->message);
	    g_error_free (error);
	    error = NULL;
	  }

	g_free (tuple);
	g_free (path);
    }
}

void
plugins_shutdown (void)
{
#if 0
    GError *error = NULL;
    gint n;

    n = 0;
    while (plugin_type_names[n])
     {
	n++;
     }
#endif
}

/* Tracklist exporting */
void
container_h_tuple_free (gpointer data)
{
    ContainerHTuple *tuple = (ContainerHTuple*)data;

    g_free (tuple->suffix);
    g_free (tuple->name);
    g_free (tuple->desc);
    g_free (tuple);
}

static void
for_each_container (gpointer key, gpointer value, GList **containers)
{
    BmpPluginContainerInterface *container = (BmpPluginContainerInterface*)value;

    if (BMP_PLUGIN_CONTAINER_GET_IFACE (container)->get_export_data != NULL) 
      {
	ContainerHTuple *tuple = g_new0 (ContainerHTuple,1);	
	const gchar** info;

	info = bmp_plugin_container_get_export_data (container);

	tuple->name = g_strdup (key);
	tuple->desc = g_strdup (info[0]);
	tuple->suffix = g_strdup (info[1]);
    
	*containers = g_list_append (*containers, tuple); 
      }
}

GList*
plugins_find_containers (void)
{
  GList *containers = NULL;

  g_hash_table_foreach (plugins_table[BMP_PLUGIN_TYPE_CONTAINER], (GHFunc)for_each_container, &containers);

  return containers;
}
