#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#else
#include <glib.h>
#endif

#include <glib/gstdio.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "main.h"

#include <taglib/tag_c.h>

#include <libhrel/types.h>
#include <libhrel/relation.h>

#include "bmpx/util.h"
#include "bmpx/file_utils.h"
#include "bmpx/metadata.h"
#include "bmpx/uri.h"

#ifdef HAVE_GUI
#include "ui_splash.h"
#include "ui_dialog_library.h"
#endif

#include "lcmfs.h"

G_DEFINE_TYPE (BmpMetadataCache, bmp_metadata_cache, G_TYPE_OBJECT)

enum {
    BMP_METADATA_CACHE_SIGNAL_FILE_ADDED,

    BMP_METADATA_N_SIGNALS
};
static guint signals[BMP_METADATA_N_SIGNALS] = {0};

enum {
    LIST_ARTIST,
    LIST_ALBUM
};

struct _BmpMetadataCachePrivate {
        gboolean dispose_has_run;

	FILE *fh;

	HRelation   *relation;
};

struct _BmpTagDefine {
    gchar *title;
    gchar *id;
    GType  type;

    gint   default_int;
    gchar* default_char;
};
typedef struct _BmpTagDefine BmpTagDefine;

/* We're using GST-style tag identifiers here */
static BmpTagDefine bmp_tag_defines[] = {
    {"Location",          "location",       G_TYPE_STRING, 0, "(Unknown Location)"},
    {"Title",             "title",          G_TYPE_STRING, 0, "(Unknown Title)"},
    {"Artist",            "artist",         G_TYPE_STRING, 0, "(Unknown Artist)"},
    {"Album",             "album",          G_TYPE_STRING, 0, "(Unknown Album)"},
    {"Genre",             "genre",          G_TYPE_STRING, 0, "(Unknown Genre)"},
    {"Comment",           "comment",        G_TYPE_STRING, 0, "(Unknown Comment)"},
    {"Time",              "time",           G_TYPE_INT,	   0, ""},
    {"Track",             "track-number",   G_TYPE_INT,	   0, ""},
    {"Rating",            "rating",         G_TYPE_INT,	   0, ""},
    {"Date",              "date",           G_TYPE_INT,	   0, ""},
    {"MTIME",		  "mtime",	    G_TYPE_INT,	   0, ""},
    {"Bitrate",           "bitrate",        G_TYPE_INT,	   0, ""},
    {"Samplerate",        "samplerate",     G_TYPE_INT,	   0, ""},
    {"Play count",        "count",	    G_TYPE_INT,	   0, ""},
};

G_LOCK_DEFINE(METADATA);

static GHashTable*
tuple_to_hash_table (HTuple *tuple)
{
  GHashTable *array;
  GValue *value;
  HTupleElem *elmt;
  const gchar *key_name;
  gint n;

  array = g_hash_table_new (g_str_hash, g_str_equal); 

  for (n = 0; n < G_N_ELEMENTS(bmp_tag_defines); n++)
    {
      key_name = 
	  bmp_metadata_get_id_static (n); 
      elmt = 
	  h_tuple_get (tuple, key_name); 

      value = g_new0 (GValue,1);
      g_value_init (value, elmt->type);

      if (elmt->type == G_TYPE_INT)
  	g_value_set_int (value, GPOINTER_TO_INT(elmt->value));
      else if (elmt->type == G_TYPE_STRING)
	g_value_set_string (value, (gchar*)(elmt->value));

      g_hash_table_insert (array, g_strdup(key_name), value); 
    }

  return array;
}

GHashTable*
bmp_metadata_cache_get_hash_value_table (BmpMetadataCache *self, const gchar *uri)
{
  HTuple *tuple;

  tuple = bmp_metadata_cache_get_metadata (self, uri);

  if (!tuple)
    return NULL;

  return tuple_to_hash_table (tuple);
}

gchar*
bmp_metadata_get_id (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].id);
}

GType
bmp_metadata_get_type (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].type;
}


const gchar*
bmp_metadata_get_id_static (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].id;
}

gchar*
bmp_metadata_get_title (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].title);
}

const gchar*
bmp_metadata_get_default_char (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].default_char);
}

gint
bmp_metadata_get_default_int (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].default_int;
}

HTuple*
bmp_metadata_tuple_create_default (void)
{
    HTuple *tuple;
    gint    n;

    tuple = h_tuple_new (); 
    for (n = 0; n < G_N_ELEMENTS(bmp_tag_defines); n++)
      {
	GType type = bmp_metadata_get_type (n);
	h_tuple_insert (tuple,
			bmp_metadata_get_id_static (n), 
			type, 
			(type == G_TYPE_INT) ? GINT_TO_POINTER (bmp_metadata_get_default_int (n)) 
			: (gpointer)bmp_metadata_get_default_char (n)); 
      }
    return tuple;
}

static HTuple*
metadata_get_taglib (BmpMetadataCache *self, const gchar *uri) 
{
    TagLib_File			  *file = NULL;
    TagLib_Tag			  *tag  = NULL;
    TagLib_AudioProperties const  *audio = NULL;
    HTuple			  *tuple;
    GURI			  *g_uri;
    struct stat			   file_stat;
    const gchar			  *_aux;
    gchar			  *filename, 
				  *_aux_check;

    g_uri = bmp_uri_new (uri);
    bmp_uri_unescape (g_uri);
    filename = g_uri->path;

    if (!filename)
      {
	bmp_uri_delete (g_uri);
	return NULL;
      }

    tuple = bmp_metadata_tuple_create_default ();

    stat (filename, &file_stat);

    h_tuple_insert (tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_MTIME),    
		    G_TYPE_INT,    
		    GINT_TO_POINTER(file_stat.st_mtime));

    h_tuple_insert (tuple, 
		    bmp_metadata_get_id_static (BMP_DATUM_LOCATION), 
		    G_TYPE_STRING, 
		    (gpointer)uri); 

    file = taglib_file_new (filename);

    if (file)
      {
         tag = taglib_file_tag (file);
         if (!tag)
           {
              g_message ("%s: no metadata", filename);
	      taglib_file_free (file);
	      return NULL;
           }

         audio = taglib_file_audioproperties (file);

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_TRACK),    
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_tag_track (tag))); 

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_DATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_tag_year  (tag))); 

	 /* G_TYPE_STRING metadata */
	 _aux = taglib_tag_artist (tag);
	 _aux_check = _aux ? g_strstrip(g_strdup(_aux)) : NULL; 
	 if (_aux && strlen(_aux_check))
	   {
	      h_tuple_insert (tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_ARTIST), 
			      G_TYPE_STRING, 
			      (gpointer)_aux); 
	   }
	 g_free (_aux_check);

	 _aux = taglib_tag_album (tag);
	 _aux_check = _aux ? g_strstrip(g_strdup(_aux)) : NULL; 
	 if (_aux && strlen(_aux_check))
	   {
	      h_tuple_insert (tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_ALBUM), 
			      G_TYPE_STRING, 
			      (gpointer)_aux); 
	   }
	 g_free (_aux_check);

	 _aux = taglib_tag_title (tag);
	 _aux_check = _aux ? g_strstrip(g_strdup(_aux)) : NULL; 
	 if (_aux && strlen(_aux_check))
	   {
	      h_tuple_insert (tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_TITLE), 
			      G_TYPE_STRING, 
			      (gpointer)_aux); 
	   }
	 g_free (_aux_check);

	 _aux = taglib_tag_genre (tag);
	 _aux_check = _aux ? g_strstrip(g_strdup(_aux)) : NULL; 
	 if (_aux && strlen(_aux_check))
	   {
	      h_tuple_insert (tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_GENRE), 
			      G_TYPE_STRING, 
			      (gpointer)_aux); 
	   }
	 g_free (_aux_check);

	 _aux = taglib_tag_comment (tag);
	 _aux_check = _aux ? g_strstrip(g_strdup(_aux)) : NULL; 
	 if (_aux && strlen(_aux_check))
	   {
	      h_tuple_insert (tuple, 
			      bmp_metadata_get_id_static (BMP_DATUM_COMMENT), 
			      G_TYPE_STRING, 
			      (gpointer)_aux); 
	   }
	 g_free (_aux_check);

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_RATING),    
			 G_TYPE_INT,    
			 GINT_TO_POINTER(0)); 

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_COUNT),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(0));

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_BITRATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_bitrate(audio)));

	 h_tuple_insert (tuple, 
			 bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_samplerate(audio)));

	 h_tuple_insert (tuple, 
		         bmp_metadata_get_id_static (BMP_DATUM_TIME),     
			 G_TYPE_INT,    
			 GINT_TO_POINTER(taglib_audioproperties_length (audio))); 


         taglib_tag_free_strings ();
         taglib_file_free (file);
     }

    bmp_uri_delete (g_uri);
    return tuple;
}

/* Media Cache functions */
static void
bmp_metadata_cache_init (BmpMetadataCache *self)
{
        self->private = g_new (BmpMetadataCachePrivate, 1);

        self->private->dispose_has_run = FALSE; 

	h_type_init ();

	self->private->relation = 
	    h_relation_base_new ("track",
                               "track has location, title, artist, album, genre, comment, track, rating, date, mtime, bitrate, samplerate, count",
                               bmp_metadata_get_id_static (BMP_DATUM_LOCATION),	  bmp_metadata_get_type (BMP_DATUM_LOCATION),
                               bmp_metadata_get_id_static (BMP_DATUM_TITLE),	  bmp_metadata_get_type (BMP_DATUM_TITLE),
                               bmp_metadata_get_id_static (BMP_DATUM_ARTIST),	  bmp_metadata_get_type (BMP_DATUM_ARTIST),
                               bmp_metadata_get_id_static (BMP_DATUM_ALBUM),	  bmp_metadata_get_type (BMP_DATUM_ALBUM),
                               bmp_metadata_get_id_static (BMP_DATUM_GENRE),	  bmp_metadata_get_type (BMP_DATUM_GENRE),
                               bmp_metadata_get_id_static (BMP_DATUM_COMMENT),	  bmp_metadata_get_type (BMP_DATUM_COMMENT),
                               bmp_metadata_get_id_static (BMP_DATUM_TIME),	  bmp_metadata_get_type (BMP_DATUM_TIME),
                               bmp_metadata_get_id_static (BMP_DATUM_TRACK),	  bmp_metadata_get_type (BMP_DATUM_TRACK),
                               bmp_metadata_get_id_static (BMP_DATUM_RATING),	  bmp_metadata_get_type (BMP_DATUM_RATING),
                               bmp_metadata_get_id_static (BMP_DATUM_DATE),	  bmp_metadata_get_type (BMP_DATUM_DATE),
                               bmp_metadata_get_id_static (BMP_DATUM_MTIME),	  bmp_metadata_get_type (BMP_DATUM_MTIME),
                               bmp_metadata_get_id_static (BMP_DATUM_BITRATE),     bmp_metadata_get_type (BMP_DATUM_BITRATE),
                               bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),  bmp_metadata_get_type (BMP_DATUM_SAMPLERATE),
                               bmp_metadata_get_id_static (BMP_DATUM_COUNT),	  bmp_metadata_get_type (BMP_DATUM_COUNT),
                               NULL,
                               bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
                               NULL);

	g_message("%s\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static void
bmp_metadata_cache_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_metadata_cache_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpMetadataCacheClass *klass;
                GObjectClass *parent_class;
                klass = BMP_METADATA_CACHE_CLASS (g_type_class_peek (BMP_TYPE_METADATA_CACHE));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_metadata_cache_dispose (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        if (self->private->dispose_has_run)
	  return;

        self->private->dispose_has_run = TRUE;

	lcmfs_close ();
	g_message("%s\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_finalize (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        g_free (self->private);
}

BmpMetadataCache*
bmp_metadata_cache_new (void)
{
    BmpMetadataCache *self;
    gchar *lcmfs_base_path;

    self = g_object_new (bmp_metadata_cache_get_type (), NULL);

#ifdef HAVE_GUI
    library_metadata_hookup (self);
#endif

    lcmfs_base_path = g_build_filename (BMP_PATH_USER_DIR, "lcmfs", NULL);

    lcmfs_init ();
    lcmfs_open (lcmfs_base_path);
    g_free (lcmfs_base_path);

#ifdef HAVE_GUI
    /* Hook splash stuff.  We need to do it here, else 'self' is not
       initialized, or its too late. */
    bmp_splash_start_progress_bar (self);
#endif
    
    lcmfs_read (self);

    return self;
}

static void
bmp_metadata_cache_class_init (BmpMetadataCacheClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = bmp_metadata_cache_set_property;
        gobject_class->get_property = bmp_metadata_cache_get_property;
        gobject_class->dispose = bmp_metadata_cache_dispose;
        gobject_class->finalize = bmp_metadata_cache_finalize;
        gobject_class->constructor = bmp_metadata_cache_constructor;

        signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED] =
	    g_signal_new ("file-added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, file_added),
                      NULL, NULL, g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE, 1, G_TYPE_POINTER);

}

/* Private API */
void
_bmp_metadata_cache_tuple_inserted (BmpMetadataCache *self, HTuple *tuple)
{
    h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation), tuple);
    g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED], 0, tuple); 
}

/* Public API */
HTuple*
bmp_metadata_cache_get_metadata (BmpMetadataCache *self, const gchar *uri)
{
    HTuple *tuple = NULL;

//    G_LOCK(METADATA);

    tuple = h_relation_base_get_tuplev (H_RELATION_BASE(self->private->relation), 
					bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
				        (gpointer)uri);
 
    if (!tuple)
    {
	if (!g_ascii_strncasecmp("file:",uri,5))
	  {
		tuple = metadata_get_taglib (self, uri);
#ifdef HAVE_XINE
		if (!tuple) tuple = bmp_play_get_metadata_xine (bmp_play, uri);
#endif
	  }
#ifdef HAVE_XINE
	else tuple = bmp_play_get_metadata_xine (bmp_play, uri);
#endif

	if (tuple)
	  {
	    if (lcmfs_record_add (tuple))
	      {
		h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation), tuple);
		g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED], 0, tuple); 
	      }
	  }
     }

//    G_UNLOCK(METADATA);

    return tuple; 
}

gchar*
bmp_metadata_cache_get_titlestring (BmpMetadataCache *self, const gchar *uri)
{
    GString *string_out = NULL;
    GString *string_field = NULL;
    HTuple *tuple;
    gchar *field;
    const gchar *format;
    gint n;
    gboolean hit = FALSE;

    g_return_val_if_fail (uri != NULL, NULL);
    
    /* Use existing cache rather than lookup for http://, as it should already be
     * there. Also, we do not want the playlist expose code to do lookups, else it
     * "stalls" while waiting for the http fetch. */
    if (!g_ascii_strncasecmp ("http:", uri, 5))
      {
	tuple = h_relation_base_get_tuplev (H_RELATION_BASE(self->private->relation), 
					    bmp_metadata_get_id_static (BMP_METADATA_PKEY), 
					    (gpointer)uri);
      }
    else
      {
	tuple = bmp_metadata_cache_get_metadata (self, uri);
      }

    if (!tuple)
      return g_strdup("");

    format = xcs_key_fetch (xcs, "bmp", "titlestring-format");
    string_out = g_string_new ("");

    while (*format) {

	/* Begin sequence */
	if ((*format == '%') && (*(format+1) != 0) && (*(format+1) == '{'))
	  {
	    string_field =
		g_string_new ("");
	    format += 2;

	    while (*format && (*format != '}') && (*format != ' '))
	      {
		g_string_append_c (string_field, *format);
		format++;
	      }

	    field =
		g_string_free (string_field, FALSE);

	    if (!*format)
	      {
	        g_free (field);
		h_tuple_free (tuple);
		return g_string_free (string_out, FALSE);
	      }

	    /* Append metadata field */
	    for (n = 0; n < BMP_DATUM_NONE; n++)
	      {
		if (!g_ascii_strcasecmp (field, bmp_metadata_get_id_static(n)))
                  {
		    HTupleElem *elem = NULL;

		    elem = h_tuple_get (tuple, field);

		    if (elem)
		      {
			gchar *value = NULL; 

			if (elem->type == G_TYPE_INT)
			  {
			    value = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
			  }
			else if (elem->type == G_TYPE_STRING)
			  {
			    if (strcmp (elem->value, bmp_metadata_get_default_char (n)))
			      value = g_strdup (elem->value);
			    else
			      value = NULL;
			  }
			
			if (value)
			  {
			    g_strstrip (value);
			    if (strlen(value)>0 && strcmp(value, "0") != 0)
			      {
				hit = TRUE;
				g_string_append (string_out, value);
			      }
			    g_free (value);
			  }
		      }
		    break;
		  }
	      }

	    g_free (field);
	  }
	else
	  {
	    g_string_append_c (string_out, *format);
	  }
	format++;
    }

    h_tuple_free (tuple);

    if (!hit)
      {
	gchar *filename, *basename;

	g_string_free (string_out, TRUE);

	if (!g_ascii_strncasecmp("file:",uri,5))
	  {
	    filename = g_filename_from_uri (uri, NULL, NULL);

	    if (filename)
	      {
		basename = g_path_get_basename (filename);
		g_free (filename);
		if (basename)
		  return basename;
	      }
	  }
	
	return g_strdup (uri);
      }
    else
	return g_string_free (string_out, FALSE);
}

Query*
bmp_metadata_cache_query_construct (BmpMetadataCache *self, gint n_atoms, ...)
{
    Query      *query;
    GType	type;
    va_list	args;
    gint	n;
    gboolean	fatal_error = FALSE;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE(self), NULL);

    /* Initialize query */
    query = g_new0 (Query,1);
    query->query_atoms = NULL;

    va_start (args, n_atoms);

    for (n = 0; n < n_atoms; n++) {

        QueryAtom *query_atom = g_new0 (QueryAtom,1);

        query_atom->tag_id =
            va_arg (args, gint);

        query_atom->exact_match =
            va_arg (args, gboolean);
      
	query_atom->match_value = g_new0 (GValue,1); 

	type = bmp_metadata_get_type (query_atom->tag_id);
	g_value_init (query_atom->match_value, type); 

	switch (type)
	  {
	    case G_TYPE_INT:
	      {
		  g_value_set_int (query_atom->match_value, va_arg (args, gint));
	      }
	      break;

	    case G_TYPE_STRING:
	      {
		  g_value_set_string (query_atom->match_value, va_arg (args, gchar*));
	      }
	      break;

	    default:
	      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       query_atom->tag_id,
		       (gulong)type);

		fatal_error = TRUE;
		break;
	      }
	  }

	if (fatal_error) break;

        query->query_atoms = g_list_append (query->query_atoms, query_atom);
    }

    va_end (args);

    // FIXME: Free query on error
    return fatal_error ? NULL : query;
}

gboolean
query_predicate_func (HTuple	 *tuple,
		      Query	 *query) 
{
  GList *_iter_list = query->query_atoms;
  gboolean hit = TRUE;

  if (g_ascii_strncasecmp("file:", h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY))->value,5)) 
    return FALSE;

  for ( ; _iter_list; _iter_list = _iter_list->next)
    {
      QueryAtom  *atom = (QueryAtom*)_iter_list->data;
      HTupleElem *elem = h_tuple_get (tuple, bmp_metadata_get_id_static (atom->tag_id)); 

      switch (elem->type)
	{
	    case G_TYPE_INT:
	      {
		  if (GPOINTER_TO_INT(elem->value) != g_value_get_int (atom->match_value))
		      hit = FALSE;
	      }
	      break;

	    case G_TYPE_STRING:
	      {
		  if (!atom->exact_match)
		    {
		      if (!match_keys((gchar*)elem->value, 
				g_value_get_string (atom->match_value))) 
		      hit = FALSE;
		    }
		  else
		    {
		      if (g_utf8_collate((gchar*)elem->value, 
					 g_value_get_string (atom->match_value))) 
		      hit = FALSE;
		    }
	      }
	      break;

	    default:
	      {
		g_log (G_LOG_DOMAIN, 
		       G_LOG_LEVEL_CRITICAL, 
		       "%s(%d): System inconsistency: Metadatum %d is not of type %ld", 
		       __FILE__, 
		       __LINE__, 
		       atom->tag_id,
		       (gulong)elem->type);
		break;
	      }
      }

      if (!hit)
	break;
    }

    return hit;
}

HRelation*
bmp_metadata_cache_query_run (BmpMetadataCache *self, Query *query) 
{
    HRelation *relation;

    g_return_val_if_fail (BMP_IS_METADATA_CACHE (self), NULL);

    relation = h_relation_restrict (self->private->relation,
				    (HTuplePredFunc)query_predicate_func,
				    query);	

    return relation;
}
