/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2003  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 2000-2003  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <bmpx/metadata.h>

#include "main.h"
#include "lcmfs.h"
#include "ui_util.h"
#include "ui_splash.h"

static GtkWidget *splash_window = NULL;
static GdkPixbuf *splash_pixbuf = NULL;

static gint splash_width = 0;
static gint splash_height = 0;
static gint splash_progress = 0;

static gpointer obj_instance = NULL;
static gulong handler_id = 0;

#define BAR_X		(10)
#define BAR_Y		(splash_height - 38)
#define BAR_WIDTH	(splash_width - 20)
#define BAR_HEIGHT	(15)

static gboolean
on_splash_window_expose (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  cairo_t *cr;
  GdkColor color;
  gint progress;

  if (splash_progress == 100)
    progress = BAR_WIDTH - 4;
  else if (splash_progress == 0)
    progress = 0;
  else
    progress = ((BAR_WIDTH - 4) * splash_progress) / 100;

  cr = gdk_cairo_create (widget->window);
  
  gdk_cairo_set_source_pixbuf (cr, splash_pixbuf, 0, 0);

  if (event)
    {
      cairo_reset_clip (cr);
      gdk_cairo_region (cr, event->region);
      cairo_clip (cr);
    }

  cairo_paint (cr);

  color.red   = 65535;
  color.green = 65535;
  color.blue  = 65535;

  draw_rounded_rectangle (cr,
			  BAR_X, 
			  BAR_Y,
			  BAR_WIDTH,
			  BAR_HEIGHT,
			  &color,
			  0.8,
			  1.0,
			  FALSE, 4); 

  draw_rounded_rectangle (cr,
			  BAR_X+2, 
			  BAR_Y+2,
			  progress,
			  BAR_HEIGHT-4,
			  &color,
			  0.6,
			  1.0,
			  TRUE, 2); 

  cairo_destroy (cr);
  
  return FALSE;
}

static void
update_splash_progress (gpointer data)
{
  static gint count = 0;
  gint lcmfs_index_size, progress;
  
  if (!splash_window)
    return;

  lcmfs_index_size = lcmfs_get_index_size ()-1;
 
  if (lcmfs_index_size > 0) 
    progress = (count++ * 100) / (lcmfs_index_size);
  else
    progress = 0;

  if (progress > 100)
    progress = 100;
  else if (progress < 0)
    progress = 0;

  if (progress != splash_progress)
    {
      splash_progress = progress;

      gtk_widget_queue_draw_area (splash_window,
				  BAR_X, BAR_Y,
				  BAR_WIDTH, BAR_HEIGHT);
    }
}

void
bmp_splash_window_show(void)
{
  GdkPixmap *mask_pixmap_window1 = NULL,
	    *mask_pixmap_window2 = NULL;
  GdkBitmap *mask_bitmap_window1 = NULL,
	    *mask_bitmap_window2 = NULL;
  gchar *filename;

  if (splash_window)
    return;

  splash_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_type_hint(GTK_WINDOW(splash_window),
			   GDK_WINDOW_TYPE_HINT_DIALOG);

  filename = DATA_DIR G_DIR_SEPARATOR_S "images" G_DIR_SEPARATOR_S "splash.png";
  splash_pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
  splash_width = gdk_pixbuf_get_width (splash_pixbuf);
  splash_height = gdk_pixbuf_get_height (splash_pixbuf);

  gtk_widget_set_size_request (GTK_WIDGET (splash_window),
			       splash_width,
			       splash_height);

  gdk_pixbuf_render_pixmap_and_mask (splash_pixbuf,
				     &mask_pixmap_window1,
				     &mask_bitmap_window1,
				     0);

  gdk_pixbuf_render_pixmap_and_mask (splash_pixbuf,
				     &mask_pixmap_window2,
				     &mask_bitmap_window2,
				     128);

  gtk_widget_add_events (splash_window, GDK_ALL_EVENTS_MASK);

  g_signal_connect (splash_window,
		    "expose_event",
		    G_CALLBACK(on_splash_window_expose),
		    NULL);
  
  gtk_widget_set_app_paintable (splash_window, TRUE);

  gtk_widget_shape_combine_mask (GTK_WIDGET (splash_window), mask_bitmap_window2, 0, 0);

  gtk_window_set_decorated (GTK_WINDOW (splash_window), FALSE);
  bmp_window_set_icon_list (GTK_WIDGET(splash_window), "player");
  gtk_window_set_title (GTK_WINDOW (splash_window), "BMPx Starting...");
  gtk_window_set_position (GTK_WINDOW (splash_window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (splash_window), FALSE);

  gtk_widget_show_all(splash_window);
  gdk_window_raise (splash_window->window);
}

void
bmp_splash_start_progress_bar (gpointer instance)
{
  obj_instance = instance;
  handler_id = g_signal_connect (G_OBJECT(instance),
				 "file-added",
				 G_CALLBACK(update_splash_progress),
				 NULL);
}

void
bmp_splash_window_hide (void)
{
  if (splash_window)
    {
      gint i;
      
      g_signal_handler_disconnect (G_OBJECT(obj_instance),
				   handler_id);

      while (gtk_events_pending ())
	gtk_main_iteration ();

      /* Make sure the progress bar reach 100 if there is no entries
       * in the cache */
      splash_progress = 100;
      gtk_widget_queue_draw (splash_window);
      for (i = 0; i < 100; i++)
	{
	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	  g_usleep (10);
	}

      gtk_widget_hide (splash_window);
      gtk_widget_destroy (splash_window);
      g_object_unref (splash_pixbuf);
      splash_window = NULL;
    }
}
  
